# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from threading import Thread

import gi
gi.require_version('GLib', '2.0')
gi.require_version('Hinawa', '3.0')
from gi.repository import GLib, Hinawa

from hinawa_utils.dg00x.config_rom_parser import Dg00xConfigRomParser

__all__ = ['Dg00xUnit']


class Dg00xUnit(Hinawa.SndDg00x):
    __BASE_ADDR = 0xffffe0000000

    SUPPORTED_SAMPLING_RATES = (44100, 48000, 88200, 96000)
    SUPPORTED_CLOCK_SOURCES = ('Internal', 'S/PDIF', 'ADAT', 'Word-clock')
    SUPPORTED_OPTICAL_INTERFACES = ('ADAT', 'S/PDIF')

    def __init__(self, path):
        super().__init__()
        self.open(path)
        if self.get_property('type') != 5:
            raise ValueError('The character device is not for Dg00x unit')

        ctx = GLib.MainContext.new()
        self.create_source().attach(ctx)
        self.__unit_dispatcher = GLib.MainLoop.new(ctx, False)
        self.__unit_th = Thread(target=lambda d: d.run(), args=(self.__unit_dispatcher, ))
        self.__unit_th.start()

        node = self.get_node()
        ctx = GLib.MainContext.new()
        node.create_source().attach(ctx)
        self.__node_dispatcher = GLib.MainLoop.new(ctx, False)
        self.__node_th = Thread(target=lambda d: d.run(), args=(self.__node_dispatcher, ))
        self.__node_th.start()

        parser = Dg00xConfigRomParser()
        info = parser.parse_rom(self.get_node().get_config_rom())
        self._model_name = info['model-name']

    def release(self):
        self.__unit_dispatcher.quit()
        self.__node_dispatcher.quit()
        self.__unit_th.join()
        self.__node_th.join()

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, trace):
        self.release()

    def _read_transaction(self, offset, size):
        req = Hinawa.FwReq()
        addr = self.__BASE_ADDR + offset
        if size == 4:
            tcode = Hinawa.FwTcode.READ_QUADLET_REQUEST
        else:
            tcode = Hinawa.FwTcode.READ_BLOCK_REQUEST
        frames = bytearray(size)
        return req.transaction(self.get_node(), tcode, addr, size, frames)

    def _write_transaction(self, offset, size):
        req = Hinawa.FwReq()
        addr = self.__BASE_ADDR + offset
        if size == 4:
            tcode = Hinawa.FwTcode.WRITE_QUADLET_REQUEST
        else:
            tcode = Hinawa.FwTcode.WRITE_BLOCK_REQUEST
        frames = bytearray(size)
        return req.transaction(self.get_node(), tcode, addr, size, frames)

    def set_clock_source(self, source):
        if source not in self.SUPPORTED_CLOCK_SOURCES:
            raise ValueError('Invalid argument for clock source.')
        data = bytearray(4)
        data[3] = self.SUPPORTED_CLOCK_SOURCES.index(source)
        self._write_transaction(0x0118, data)

    def get_clock_source(self):
        data = self._read_transaction(0x0118, 4)
        if data[3] >= len(self.SUPPORTED_CLOCK_SOURCES):
            raise OSError('Unexpected value for clock source.')
        return self.SUPPORTED_CLOCK_SOURCES[data[3]]

    def set_local_sampling_rate(self, rate):
        if rate not in self.SUPPORTED_SAMPLING_RATES:
            raise ValueError('Invalid argument for local sampling rate.')
        data = bytearray(4)
        data[3] = self.SUPPORTED_SAMPLING_RATES.index(rate)
        self._write_transaction(0x0110, data)

    def get_local_sampling_rate(self):
        data = self._read_transaction(0x0110, 4)
        if data[3] >= len(self.SUPPORTED_SAMPLING_RATES):
            raise OSError('Unexpected value for local sampling rate.')
        return self.SUPPORTED_SAMPLING_RATES[data[3]]

    def get_external_sampling_rate(self):
        data = self._read_transaction(0x0114, 4)
        if data[3] >= len(self.SUPPORTED_SAMPLING_RATES):
            raise OSError('Unexpected value for external sampling rate.')
        return self.SUPPORTED_SAMPLING_RATES[data[3]]

    def check_external_input(self):
        data = self._read_transaction(0x012c, 4)
        return data[3] > 0

    def set_opt_iface(self, mode):
        if mode not in self.SUPPORTED_OPTICAL_INTERFACES:
            raise ValueError('Invalid argument for optical interface mode.')
        data = bytearray(4)
        data[3] = self.SUPPORTED_OPTICAL_INTERFACES.index(mode)
        self._write_transaction(0x011c, data)

    def get_opt_iface(self):
        data = self._read_transaction(0x011c, 4)
        if data[3] >= len(self.SUPPORTED_OPTICAL_INTERFACES):
            raise ValueError('Unexpected value for optical interface mode.')
        return self.SUPPORTED_OPTICAL_INTERFACES[data[3]]

    def set_mixer_mode(self, mode):
        if mode > 0:
            mode = 1
        else:
            mode = 0
        data = bytearray(4)
        data[3] = mode
        self._write_transaction(0x0124, data)

    def get_mixer_mode(self):
        data = self._read_transaction(0x0124, 4)
        return (data[3] > 0)



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from struct import pack
from enum import Enum

import gi
gi.require_version('Hinawa', '3.0')
from gi.repository import Hinawa

from hinawa_utils.ta1394.general import AvcGeneral

__all__ = ['HwCmd', 'DisplayCmd', 'OptIfaceCmd', 'MicCmd', 'InputCmd',
           'OutputCmd', 'MixerCmd', 'RouteCmd', 'KnobCmd', 'SpdifResampleCmd']


class VendorCmd(Enum):
    INPUT_LIMIT = 0xe4
    MIC_POWER = 0xe5
    IO_ATTR = 0xe8
    IO_ROUTING = 0xef
    HW = 0xeb
    HP_SRC = 0xab
    MIXER_SRC1 = 0xb0
    MIXER_SRC2 = 0xb1
    MIXER_SRC3 = 0xb2
    MIXER_SRC4 = 0xb3
    OPT_IFACE_MODE = 0xf1
    DOWNGRADE = 0xf2
    SPDIF_RESAMPLE = 0xf3
    MIC_POLARITY = 0xf5
    KNOB_VALUE = 0xf6
    HW_STATUS = 0xff


class HwCmdOp(Enum):
    STREAM_MODE = 0x06
    DISPLAY_ILLUMINATE = 0x08
    DISPLAY_MODE = 0x09
    DISPLAY_TARGET = 0x0a
    DISPLAY_OVERHOLD = 0x0e
    METER_RESET = 0x0f
    CD_MODE = 0xf5


class ApogeeProtocol():
    __OUI = (0x00, 0x03, 0xdb)

    @classmethod
    def command_set(cls, fcp: Hinawa.FwFcp, cmd: VendorCmd, args: bytearray):
        data = bytearray()
        data.append(cmd.value)

        if len(args) > 0:
            data.extend(args)

        # At least, 6 bytes should be required to align to 3 quadlets.
        # Unless, target unit is freezed.
        if len(data) < 6:
            for i in range(6 - len(data)):
                data.append(0x00)

        resps = AvcGeneral.set_vendor_dependent(fcp, cls.__OUI, data)
        if resps[0] != data[0]:
            raise OSError('Unexpected value for vendor-dependent command.')

        return resps[1:]


class HwCmd():
    __NAME = 'hw'

    __STREAM_MODES = {
        '16x16':    0x00,
        '10x10':    0x01,
        '8x8':      0x02,
    }

    __DOWNGRADE_TARGETS = {
        'none':                 0x00,
        'analog-in-1/2':        0x01,
        'analog-in-3/4':        0x02,
        'analog-in-5/6':        0x03,
        'analog-in-7/8':        0x04,
        'spdif-opt-in-1/2':     0x05,
        'spdif-coax-in-1/2':    0x09,
        'spdif-coax-out-1/2':   0x0a,
        'spdif-opt-out-1/2':    0x0b,
    }

    @classmethod
    def create_cache(cls, cache: dict):
        cache[cls.__NAME] = {
            'cd':       False,
            '16bit':    'none',
        }

    @classmethod
    def get_stream_mode_labels(cls):
        return list(cls.__STREAM_MODES.keys())

    @classmethod
    def set_stream_mode(cls, fcp: Hinawa.FwFcp, mode: str):
        if mode not in cls.__STREAM_MODES:
            raise ValueError('Invalid argument for mode of stream.')
        args = bytearray(2)
        args[0] = HwCmdOp.STREAM_MODE.value
        args[1] = cls.__STREAM_MODES[mode]
        ApogeeProtocol.command_set(fcp, VendorCmd.HW, args)

    @classmethod
    def set_cd_mode(cls, cache: dict, fcp: Hinawa.FwFcp, enable: bool):
        if not isinstance(enable, bool):
            raise ValueError('Invalid argument for mode of cd.')
        val = 0x01 if enable else 0x00

        args = bytearray(2)
        args[0] = HwCmdOp.CD_MODE.value
        args[1] = val
        ApogeeProtocol.command_set(fcp, VendorCmd.HW, args)
        cache[cls.__NAME]['cd'] = enable

    @classmethod
    def get_cd_mode(cls, cache: dict):
        return cache[cls.__NAME]['cd']

    @classmethod
    def get_16bit_mode_labels(cls):
        return list(cls.__DOWNGRADE_TARGETS.keys())

    @classmethod
    def set_16bit_mode(cls, cache: dict, fcp: Hinawa.FwFcp, target: str):
        if target not in cls.__DOWNGRADE_TARGETS:
            raise ValueError('Invalid argument for target of 16bit mode.')
        args = bytearray(1)
        args[0] = cls.__DOWNGRADE_TARGETS[target]
        ApogeeProtocol.command_set(fcp, VendorCmd.DOWNGRADE, args)
        cache[cls.__NAME]['16bit'] = target

    @classmethod
    def get_16bit_mode(cls, cache: dict):
        return cache[cls.__NAME]['16bit']


class DisplayCmd():
    __NAME = 'display'

    __TARGETS = {
        'output':   0x00,
        'input':    0x01,
    }

    @classmethod
    def create_cache(cls, cache: dict):
        cache[cls.__NAME] = {
            'illuminate': False,
            'mode': False,
            'target': 'output',
            'overhold': False,
        }

    @classmethod
    def get_target_labels(cls):
        return list(cls.__TARGETS.keys())

    @classmethod
    def __command_set(cls, fcp, op, val):
        args = bytearray(2)
        args[0] = op.value
        args[1] = val
        return ApogeeProtocol.command_set(fcp, VendorCmd.HW, args)

    @classmethod
    def set_illuminate(cls, cache: dict, fcp: Hinawa.FwFcp, enable: bool):
        if not isinstance(enable, bool):
            raise ValueError('Invalid argument for illuminate of display.')
        val = 0x01 if enable else 0x00
        cls.__command_set(fcp, HwCmdOp.DISPLAY_ILLUMINATE, val)
        cache[cls.__NAME]['illuminate'] = enable

    @classmethod
    def get_illuminate(cls, cache: dict):
        return cache[cls.__NAME]['illuminate']

    @classmethod
    def set_mode(cls, cache: dict, fcp: Hinawa.FwFcp, enable: bool):
        if not isinstance(enable, bool):
            raise ValueError('Invalid argument for mode of display.')
        val = 0x01 if enable else 0x00
        cls.__command_set(fcp, HwCmdOp.DISPLAY_MODE, val)
        cache[cls.__NAME]['mode'] = enable

    @classmethod
    def get_mode(cls, cache: dict):
        return cache[cls.__NAME]['mode']

    @classmethod
    def set_target(cls, cache: dict, fcp: Hinawa.FwFcp, target: str):
        if target not in cls.__TARGETS:
            raise ValueError('Invalid argument for target of display.')
        val = cls.__TARGETS[target]
        cls.__command_set(fcp, HwCmdOp.DISPLAY_TARGET, val)
        cache[cls.__NAME]['target'] = target

    @classmethod
    def get_target(cls, cache: dict):
        return cache[cls.__NAME]['target']

    @classmethod
    def set_overhold(cls, cache: dict, fcp: Hinawa.FwFcp, enable: bool):
        if not isinstance(enable, bool):
            raise ValueError('Invalid argument for overhold of display.')
        val = 0x01 if enable else 0x00
        cls.__command_set(fcp, HwCmdOp.DISPLAY_OVERHOLD, val)
        cache[cls.__NAME]['overhold'] = enable

    @classmethod
    def get_overhold(cls, cache: dict):
        return cache[cls.__NAME]['overhold']

    @classmethod
    def reset_meter(cls, fcp: Hinawa.FwFcp):
        cls.__command_set(fcp, HwCmdOp.METER_RESET, 0x00)


class OptIfaceCmd():
    __NAME = 'opt-iface'

    __TARGETS = {
        'output':   0x00,
        'input':    0x00,
    }
    __MODES = {
        'S/PDIF':       0x00,
        'ADAT/SMUX':    0x01,
    }

    @classmethod
    def create_cache(cls, cache: dict):
        cache[cls.__NAME] = {}
        for target in cls.__TARGETS:
            cache[cls.__NAME][target] = 'S/PDIF'

    @classmethod
    def get_target_labels(cls):
        return list(cls.__TARGETS.keys())

    @classmethod
    def get_mode_labels(cls):
        return list(cls.__MODES.keys())

    @classmethod
    def set_mode(cls, cache: dict, fcp: Hinawa.FwFcp, target: str, mode: str):
        if target not in cls.__TARGETS:
            raise ValueError('Invalid argument for target of optical iface.')
        if mode not in cls.__MODES:
            raise ValueError('Invalid argument for mode of optical iface.')
        args = bytearray(2)
        args[0] = cls.__TARGETS[target]
        args[1] = cls.__MODES[mode]
        ApogeeProtocol.command_set(fcp, VendorCmd.OPT_IFACE_MODE, args)
        cache[cls.__NAME][target] = mode

    @classmethod
    def get_mode(cls, cache: dict, target: str):
        if target not in cls.__TARGETS:
            raise ValueError('Invalid argument for optical iface.')
        return cache[cls.__NAME][target]


class MicCmd():
    __NAME = 'mic'

    __TARGETS = {
        'mic-1':    0x00,
        'mic-2':    0x01,
        'mic-3':    0x02,
        'mic-4':    0x03,
    }

    @classmethod
    def create_cache(cls, cache: dict):
        cache[cls.__NAME] = {}
        for item in ('power', 'polarity'):
            cache[cls.__NAME][item] = {}
            for target in cls.__TARGETS:
                cache[cls.__NAME][item][target] = False

    @classmethod
    def __command_set(cls, fcp, cmd, target, val):
        if target not in cls.__TARGETS:
            raise ValueError('Invalid argument for mic.')
        args = bytearray(2)
        args[0] = cls.__TARGETS[target]
        args[1] = val
        ApogeeProtocol.command_set(fcp, cmd, args)

    @classmethod
    def get_mic_labels(cls):
        return list(cls.__TARGETS.keys())

    @classmethod
    def set_power(cls, cache: dict, fcp: Hinawa.FwFcp, target: str, enable: bool):
        val = 0x01 if enable else 0x00
        cls.__command_set(fcp, VendorCmd.MIC_POWER, target, val)
        cache[cls.__NAME]['power'][target] = enable

    @classmethod
    def get_power(cls, cache: dict, target: str):
        if target not in MicCmd.get_mic_labels():
            raise ValueError('Invalid argument for mic.')
        return cache[cls.__NAME]['power'][target]

    @classmethod
    def set_polarity(cls, cache: dict, fcp: Hinawa.FwFcp, target: str,
                     invert: bool):
        val = 0x01 if invert else 0x00
        cls.__command_set(fcp, VendorCmd.MIC_POLARITY, target, val)
        cache[cls.__NAME]['polarity'][target] = invert

    @classmethod
    def get_polarity(cls, cache: dict, target: str):
        if target not in MicCmd.get_mic_labels():
            raise ValueError('Invalid argument for mic.')
        return cache[cls.__NAME]['polarity'][target]


class InputCmd():
    __NAME = 'input'

    __TARGETS = {
        'analog-1': 0x00,
        'analog-2': 0x01,
        'analog-3': 0x02,
        'analog-4': 0x03,
        'analog-5': 0x04,
        'analog-6': 0x05,
        'analog-7': 0x06,
        'analog-8': 0x07,
    }

    __ATTRS = {
        '+4dB':     0x00,
        '-10dB':    0x01,
    }

    @classmethod
    def create_cache(cls, cache: dict):
        cache[cls.__NAME] = {
            'soft-limit': {},
            'attr': {},
        }
        for target in cls.__TARGETS:
            cache[cls.__NAME]['soft-limit'][target] = False
            cache[cls.__NAME]['attr'][target] = '-10dB'

    @classmethod
    def get_in_labels(cls):
        return list(cls.__TARGETS.keys())

    @classmethod
    def get_attr_labels(cls):
        return list(cls.__ATTRS.keys())

    @classmethod
    def set_soft_limit(cls, cache: dict, fcp: Hinawa.FwFcp, target, enable):
        if target not in cls.__TARGETS:
            raise ValueError('Invalid argument for line input.')
        val = 0x01 if enable else 0x00

        args = bytearray(2)
        args[0] = cls.__TARGETS[target]
        args[1] = val
        ApogeeProtocol.command_set(fcp, VendorCmd.INPUT_LIMIT, args)
        cache[cls.__NAME]['soft-limit'][target] = enable

    @classmethod
    def get_soft_limit(cls, cache: dict, target: str):
        if target not in cls.__TARGETS:
            raise ValueError('Invalid argument for line input.')
        return cache[cls.__NAME]['soft-limit'][target]

    @classmethod
    def set_attr(cls, cache: dict, fcp: Hinawa.FwFcp, target: str, attr: str):
        if target not in cls.__TARGETS:
            raise ValueError('Invalid argument for line input.')
        if attr not in cls.__ATTRS:
            raise ValueError('Invalid argument for in attenuation.')

        args = bytearray(3)
        args[0] = cls.__TARGETS[target]
        args[1] = 0x01  # input.
        args[2] = cls.__ATTRS[attr]
        ApogeeProtocol.command_set(fcp, VendorCmd.IO_ATTR, args)
        cache[cls.__NAME]['attr'][target] = attr

    @classmethod
    def get_attr(cls, cache: dict, target: str):
        if target not in cls.__TARGETS:
            raise ValueError('Invalid argument for line input.')
        return cache[cls.__NAME]['attr'][target]


class OutputCmd():
    __NAME = 'output'

    __TARGETS = {
        'analog-1': 0x00,
        'analog-2': 0x01,
        'analog-3': 0x02,
        'analog-4': 0x03,
        'analog-5': 0x04,
        'analog-6': 0x05,
        'analog-7': 0x06,
        'analog-8': 0x07,
    }

    __ATTRS = {
        '+4dB':     0x00,
        '-10dB':    0x01,
    }

    @classmethod
    def create_cache(cls, cache: dict):
        cache[cls.__NAME] = {
            'attr': {},
        }
        for target in cls.__TARGETS:
            cache[cls.__NAME]['attr'][target] = '-10dB'

    @classmethod
    def get_target_labels(cls):
        return list(cls.__TARGETS.keys())

    @classmethod
    def get_attr_labels(cls):
        return list(cls.__ATTRS.keys())

    @classmethod
    def set_attr(cls, cache: dict, fcp: Hinawa.FwFcp, target: str, attr: str):
        if target not in cls.__TARGETS:
            raise ValueError('Invalid argument for line output.')
        if attr not in cls.__ATTRS:
            raise ValueError('Invalid argument for out attenuation.')

        args = bytearray(3)
        args[0] = cls.__TARGETS[target]
        args[1] = 0x00    # output.
        args[2] = cls.__ATTRS[attr]
        ApogeeProtocol.command_set(fcp, VendorCmd.IO_ATTR, args)
        cache[cls.__NAME]['attr'][target] = attr

    @classmethod
    def get_attr(cls, cache: dict, target):
        if target not in cls.__TARGETS:
            raise ValueError('Invalid argument for line output.')
        return cache[cls.__NAME]['attr'][target]


class MixerCmd():
    __NAME = 'mixer'

    __TARGETS = {
        'mixer-1/2':    0x00,
        'mixer-3/4':    0x01,
    }

    __SRCS = {
        VendorCmd.MIXER_SRC1: (
            'analog-1', 'analog-2',
            'analog-3', 'analog-4',
            'analog-5', 'analog-6',
            'analog-7', 'analog-8',
            'stream-1',
        ),
        VendorCmd.MIXER_SRC2: (
            'stream-2', 'stream-3',
            'stream-4', 'stream-5',
            'stream-6', 'stream-7',
            'stream-8', 'stream-9',
            'stream-10',
        ),
        VendorCmd.MIXER_SRC3: (
            'stream-11', 'stream-12',
            'stream-13', 'stream-14',
            'stream-15', 'stream-16',
            'stream-17', 'stream-18',
            'adat-1',
        ),
        VendorCmd.MIXER_SRC4: (
            'adat-2', 'adat-3',
            'adat-4', 'adat-5',
            'adat-6', 'adat-7',
            'adat-8', 'spdif-1',
            'spdif-2',
        ),
    }

    @classmethod
    def create_cache(cls, cache: dict):
        cache[cls.__NAME] = {}
        for target in cls.__TARGETS:
            right = True
            cache[cls.__NAME][target] = {}
            for cmd, srcs in cls.__SRCS.items():
                for src in srcs:
                    db = 0.0
                    balance = 100.0 if right else 0.0
                    cache[cls.__NAME][target][src] = [db, balance]
                    right = not right

    @classmethod
    def get_target_labels(cls):
        return list(cls.__TARGETS.keys())

    @classmethod
    def get_src_labels(cls):
        srcs = []
        for cmd, sources in cls.__SRCS.items():
            srcs.extend(sources)
        return srcs

    @classmethod
    def set_src_gain(cls, cache: dict, fcp: Hinawa.FwFcp, target: str, src: str,
                     db: float, balance: float):
        if target not in cls.__TARGETS:
            raise ValueError('Invalid argument for mixer.')
        if db < -48 or db > 0:
            raise ValueError('Invalid argument for db of source.')
        if balance < 0 or 100 < balance:
            raise ValueError('Invalid argument for balance of source.')

        for cmd, srcs in cls.__SRCS.items():
            if src in srcs:
                break
        else:
            raise ValueError('Invalid argument for source of mixer.')
        data = (db, balance)

        for s in srcs:
            if s not in cache[cls.__NAME][target]:
                raise ValueError('Invalid argument for status cache.')

        args = bytearray()
        args.append(cmd.value)
        args.append(cls.__TARGETS[target])
        for s in srcs:
            if s == src:
                db, balance = data
            else:
                db, balance = cache[cls.__NAME][target][s]
            val = int(0xffff * (1 + db / 48))
            right = int(val * balance / 100)
            left = val - right
            args.extend(pack('<H', left))
            args.extend(pack('<H', right))

        ApogeeProtocol.command_set(fcp, cmd, args)
        cache[cls.__NAME][target][src] = data

    @classmethod
    def get_src_gain(cls, cache: dict, target: str, src: str):
        if target not in cls.__TARGETS:
            raise ValueError('Invalid argument for mixer.')
        for cmd, srcs in cls.__SRCS.items():
            if src in srcs:
                break
        else:
            raise ValueError('Invalid argument for source of mixer.')
        return list(cache[cls.__NAME][target][src])


class RouteCmd():
    __NAME = 'route'

    __PORTS = {
        # From external interfaces.
        'analog-1':     0x00,
        'analog-2':     0x01,
        'analog-3':     0x02,
        'analog-4':     0x03,
        'analog-5':     0x04,
        'analog-6':     0x05,
        'analog-7':     0x06,
        'analog-8':     0x07,
        # For host computer.
        'stream-1':     0x08,
        'stream-2':     0x09,
        'stream-3':     0x0a,
        'stream-4':     0x0b,
        'stream-5':     0x0c,
        'stream-6':     0x0d,
        'stream-7':     0x0e,
        'stream-8':     0x0f,
        'stream-9':     0x10,
        'stream-10':    0x11,
        'stream-11':    0x12,
        'stream-12':    0x13,
        'stream-13':    0x14,
        'stream-14':    0x15,
        'stream-15':    0x16,
        'stream-16':    0x17,
        'stream-17':    0x18,
        'stream-18':    0x19,
        # From external interfaces.
        'spdif-1':      0x1a,
        'spdif-2':      0x1b,
        'adat-1':       0x1c,
        'adat-2':       0x1d,
        'adat-3':       0x1e,
        'adat-4':       0x1f,
        'adat-5':       0x20,
        'adat-6':       0x21,
        'adat-7':       0x22,
        'adat-8':       0x23,
        # From internal multiplexers.
        'mixer-1':      0x24,
        'mixer-2':      0x25,
        'mixer-3':      0x26,
        'mixer-4':      0x27,
    }

    __OUT_TARGETS = (
        'analog-1', 'analog-2', 'analog-3', 'analog-4',
        'analog-5', 'analog-6', 'analog-7', 'analog-8',
        'spdif-1', 'spdif-2',
        'adat-1', 'adat-2', 'adat-3', 'adat-4',
        'adat-5', 'adat-6', 'adat-7', 'adat-8',
    )

    __OUT_SRCS = (
        'analog-1', 'analog-2', 'analog-3', 'analog-4',
        'analog-5', 'analog-6', 'analog-7', 'analog-8',
        'stream-1', 'stream-2', 'stream-3', 'stream-4',
        'stream-5', 'stream-6', 'stream-7', 'stream-8',
        'stream-9', 'stream-10', 'stream-11', 'stream-12',
        'stream-13', 'stream-14', 'stream-15', 'stream-16',
        'stream-17', 'stream-18',
        'spdif-1', 'spdif-2',
        'adat-1', 'adat-2', 'adat-3', 'adat-4',
        'adat-5', 'adat-6', 'adat-7', 'adat-8',
        'mixer-1', 'mixer-2',
        'mixer-3', 'mixer-4',
    )

    __CAP_TARGETS = (
        'stream-1', 'stream-2', 'stream-3', 'stream-4',
        'stream-5', 'stream-6', 'stream-7', 'stream-8',
        'stream-9', 'stream-10', 'stream-11', 'stream-12',
        'stream-13', 'stream-14', 'stream-15', 'stream-16',
        'stream-17', 'stream-18',
    )

    __CAP_SRCS = (
        'analog-1', 'analog-2', 'analog-3', 'analog-4',
        'analog-5', 'analog-6', 'analog-7', 'analog-8',
        'spdif-1', 'spdif-2',
        'adat-1', 'adat-2', 'adat-3', 'adat-4',
        'adat-5', 'adat-6', 'adat-7', 'adat-8',
    )

    __HP_TARGETS = {
        'hp-1': 0x01,
        'hp-2': 0x00,
    }

    __HP_SRCS = {
        'analog-1/2':   0x01,
        'analog-3/4':   0x03,
        'analog-5/6':   0x05,
        'analog-7/8':   0x07,
        'spdif-1/2':    0x09,
        'none':         0xff,
    }

    @classmethod
    def create_cache(cls, cache: dict):
        cache[cls.__NAME] = {
            'out':  {},
            'cap':  {},
            'hp':   {},
        }
        for i, target in enumerate(cls.__OUT_TARGETS):
            src = cls.__OUT_SRCS[i + 8]
            cache[cls.__NAME]['out'][target] = src
        for i, target in enumerate(cls.__CAP_TARGETS):
            src = cls.__CAP_SRCS[i]
            cache[cls.__NAME]['cap'][target] = src
        for i, target in enumerate(cls.__HP_TARGETS):
            src = list(cls.__HP_SRCS.keys())[i]
            cache[cls.__NAME]['hp'][target] = src

    @classmethod
    def __command_set(cls, fcp, dst, src):
        args = bytearray(2)
        args[0] = cls.__PORTS[dst]
        args[1] = cls.__PORTS[src]
        ApogeeProtocol.command_set(fcp, VendorCmd.IO_ROUTING, args)

    @classmethod
    def get_out_labels(cls):
        return cls.__OUT_TARGETS

    @classmethod
    def get_out_src_labels(cls):
        return cls.__OUT_SRCS

    @classmethod
    def set_out_src(cls, cache: dict, fcp: Hinawa.FwFcp, target: str, src: str):
        if target not in cls.__OUT_TARGETS:
            raise ValueError('Invalid argument for output.')
        if src not in cls.__OUT_SRCS:
            raise ValueError('Invalid argument for source of output.')
        cls.__command_set(fcp, target, src)
        cache[cls.__NAME]['out'][target] = src

    @classmethod
    def get_out_src(cls, cache: dict, target: str):
        if target not in cls.__OUT_TARGETS:
            raise ValueError('Invalid argument for output.')
        return cache[cls.__NAME]['out'][target]

    @classmethod
    def get_cap_labels(cls):
        return cls.__CAP_TARGETS

    @classmethod
    def get_cap_src_labels(cls):
        return cls.__CAP_SRCS

    @classmethod
    def set_cap_src(cls, cache: dict, fcp: Hinawa.FwFcp, target: str, src: str):
        if target not in cls.__CAP_TARGETS:
            raise ValueError('Invalid argument for capture.')
        if src not in cls.__CAP_SRCS:
            raise ValueError('Invalid argument for source of capture.')
        cls.__command_set(fcp, target, src)
        cache[cls.__NAME]['cap'][target] = src

    @classmethod
    def get_cap_src(cls, cache: dict, target: str):
        if target not in RouteCmd.get_cap_labels():
            raise ValueError('Invalid argument for capture.')
        return cache[cls.__NAME]['cap'][target]

    @classmethod
    def get_hp_labels(cls):
        return list(cls.__HP_TARGETS.keys())

    @classmethod
    def get_hp_src_labels(cls):
        return list(cls.__HP_SRCS.keys())

    @classmethod
    def set_hp_src(cls, cache: dict, fcp: Hinawa.FwFcp, target: str, src: str):
        if target not in cls.__HP_TARGETS:
            raise ValueError('Invalid argument for heaphone.')
        if src not in cls.__HP_SRCS:
            raise ValueError('Invalid argument for source of headphone.')
        args = bytearray(2)
        args[0] = cls.__HP_TARGETS[target]
        args[1] = cls.__HP_SRCS[src]
        ApogeeProtocol.command_set(fcp, VendorCmd.HP_SRC, args)
        cache[cls.__NAME]['hp'][target] = src

    @classmethod
    def get_hp_src(cls, cache: dict, target: str):
        if target not in cls.__HP_TARGETS:
            raise ValueError('Invalid argument for heaphone.')
        return cache[cls.__NAME]['hp'][target]


class KnobCmd():
    __IN_KNOBS = {
        # name: (value-pos, selected-value, peak-pos)
        # 0x7f - 0x00
        'mic-1':    (0, 0x00, 12),
        'mic-2':    (1, 0x08, 13),
        'mic-3':    (2, 0x10, 14),
        'mic-4':    (3, 0x18, 15),
    }

    __OUT_KNOBS = {
        # name: (value-pos, selected-mask)
        # 0x7f - 0x00
        'main':     (7, 0x01),
        'hp-1/2':   (6, 0x02),
        'hp-3/4':   (5, 0x04),
    }

    __TARGETS = {
        'main':     0x00,
        'hp-1/2':   0x01,
        'hp-3/4':   0x02,
    }

    __DB_MIN = -127.0
    __DB_MAX = 0.0

    @classmethod
    def __parse_val_from_db(cls, db):
        if db < cls.__DB_MIN or db > cls.__DB_MAX:
            raise ValueError('Invalid argument for dB.')
        return int(-db)

    @staticmethod
    def __build_val_to_db(val):
        if val > 256:
            raise ValueError('Invalid argument for linear value.')
        if val > 0x7f:
            db = (0xff - val)
        else:
            db = -val
        return float(db)

    @classmethod
    def get_knob_out_labels(cls):
        return list(cls.__TARGETS.keys())

    @classmethod
    def set_out_vol(cls, fcp: Hinawa.FwFcp, target: str, db: float):
        if target not in cls.__TARGETS:
            raise ValueError('Invalid argument for output.')
        val = cls.__parse_val_from_db(db)

        args = bytearray(2)
        args[0] = cls.__TARGETS[target]
        args[1] = val
        ApogeeProtocol.command_set(fcp, VendorCmd.KNOB_VALUE, args)

    @classmethod
    def get_states(cls, fcp: Hinawa.FwFcp):
        args = bytearray(1)
        args[0] = 0x01
        resp = ApogeeProtocol.command_set(fcp, VendorCmd.HW_STATUS, args)

        status = {}
        for label, params in cls.__IN_KNOBS.items():
            val_pos, sel_val, peak_pos = params
            db = cls.__build_val_to_db(resp[val_pos])
            status[label] = (db, bool(resp[peak_pos] > 0x00))
            if resp[4] & 0x18 == sel_val:
                status['in-knob'] = (label, resp[8])
        for label, params in cls.__OUT_KNOBS.items():
            val_pos, sel_mask = params
            db = cls.__build_val_to_db(resp[val_pos])
            status[label] = db
            if resp[4] & sel_mask:
                status['out-knob'] = (label, resp[9])
        return status


class SpdifResampleCmd():
    __NAME = 'spdif-resample'

    __IFACES = {
        'optical':  0x00,
        'coaxial':  0x01,
    }
    __DIRECTIONS = {
        'output':   0x00,
        'input':    0x01,
    }
    __RATES = {
        44100:  0x01,
        48000:  0x02,
        88200:  0x04,
        96000:  0x08,
        176400: 0x10,
        192000: 0x20,
    }

    @classmethod
    def create_cache(cls, cache: dict):
        cache[cls.__NAME] = (False, 'optical', 'output', 44100)

    @classmethod
    def get_iface_labels(cls):
        return list(cls.__IFACES.keys())

    @classmethod
    def get_direction_labels(cls):
        return list(cls.__DIRECTIONS.keys())

    @classmethod
    def get_rate_labels(cls):
        return list(cls.__RATES.keys())

    @classmethod
    def set_params(cls, cache: dict, fcp: Hinawa.FwFcp, enable: bool, iface: str,
                   direction: str, rate: int):
        if not isinstance(enable, bool):
            raise ValueError('Invalid argument for state.')
        if iface not in cls.__IFACES:
            raise ValueError('Invalid argument for iface.')
        if direction not in cls.__DIRECTIONS:
            raise ValueError('Invalid argument for direction.')
        if rate not in cls.__RATES:
            raise ValueError('Invalid argument for rate.')
        args = bytearray(4)
        args[0] = 0x01 if enable else 0x00
        args[1] = cls.__IFACES[iface]
        args[2] = cls.__DIRECTIONS[direction]
        args[3] = cls.__RATES[rate]
        ApogeeProtocol.command_set(fcp, VendorCmd.SPDIF_RESAMPLE, args)
        cache[cls.__NAME] = (enable, iface, direction, rate)

    @classmethod
    def get_params(cls, cache: dict):
        return list(cache[cls.__NAME])



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

import gi
gi.require_version('Hinawa', '3.0')
from gi.repository import Hinawa

__all__ = ['AvcGeneral', 'AvcConnection']


class AvcGeneral():
    SUBUNIT_TYPES = ('monitor', 'audio', 'printer', 'disc',
                     'tape-recorder-player', 'tuner', 'ca', 'camera',
                     'reserved', 'panel', 'bulletin-board', 'camera storate',
                     'music')
    MAXIMUM_SUBUNIT_PAGE = 0x7

    @classmethod
    def command_control(cls, fcp, cmd):
        if not isinstance(fcp, Hinawa.FwFcp):
            raise ValueError('Invalid argument for FwFcp')
        if cmd[0] != 0x00:
            raise ValueError('Invalid command code for control')
        params = [0] * 256
        params = fcp.transaction(cmd, params)
        if params[0] == 0x08:
            raise OSError('Not implemented')
        elif params[0] == 0x0a:
            raise OSError('Rejected')
        elif params[0] != 0x09:
            raise OSError('Unknown status')
        return params

    @classmethod
    def command_status(cls, fcp, cmd):
        if not isinstance(fcp, Hinawa.FwFcp):
            raise ValueError('Invalid argument for FwFcp')
        if cmd[0] != 0x01:
            raise ValueError('Invalid command code for status')
        params = [0] * 256
        params = fcp.transaction(cmd, params)
        if params[0] == 0x08:
            raise OSError('Not implemented')
        elif params[0] == 0x0a:
            raise OSError('Rejected')
        elif params[0] == 0x0b:
            raise OSError('In transition')
        elif params[0] != 0x0c:
            raise OSError('Unknown status')
        return params

    @classmethod
    def command_inquire(cls, fcp, cmd):
        if not isinstance(fcp, Hinawa.FwFcp):
            raise ValueError('Invalid argument for FwFcp')
        if cmd[0] != 0x02:
            raise ValueError('Invalid command code for inquire')
        params = [0] * 256
        params = fcp.transaction(cmd, params)
        if params[0] == 0x08:
            raise OSError('Not Implemented')
        elif params[0] != 0x0c:
            raise OSError('Unknown status')

    @classmethod
    def get_unit_info(cls, fcp):
        args = bytearray()
        args.append(0x01)
        args.append(0xff)
        args.append(0x30)
        args.append(0xff)
        args.append(0xff)
        args.append(0xff)
        args.append(0xff)
        args.append(0xff)
        params = cls.command_status(fcp, args)
        info = {}
        info['unit-type'] = params[4] >> 3
        info['unit'] = params[4] & 0x07
        info['company-id'] = (params[5], params[6], params[7])
        return info

    # NOTE: at present, this implementation doesn't support extension code.
    @classmethod
    def get_subunit_info(cls, fcp, page):
        if page > cls.MAXIMUM_SUBUNIT_PAGE:
            raise ValueError('Invalid argument for page number')
        args = bytearray()
        args.append(0x01)
        args.append(0xff)
        args.append(0x31)
        args.append(page << 4 | 0x07)
        args.append(0xff)
        args.append(0xff)
        args.append(0xff)
        args.append(0xff)
        params = cls.command_status(fcp, args)
        info = []
        for code in params[4:8]:
            if code == 0xff:
                continue
            entry = {
                'type':         cls.SUBUNIT_TYPES[code >> 3],
                'maximum-id':   code & 0x07,
            }
            info.append(entry)
        # ignoring extended_subunit_type and extended_subunit_ID
        return info

    @classmethod
    def set_vendor_dependent(cls, fcp, company_ids, deps):
        if len(company_ids) != 3:
            raise ValueError('Invalid array for company ID')
        if len(deps) == 0:
            raise ValueError('Invalid data for vendor dependent field')
        args = bytearray()
        args.append(0x00)   # Control
        args.append(0xff)   # Unit
        args.append(0x00)   # Vendor dependent command
        for b in company_ids:
            args.append(b)
        for b in deps:
            args.append(b)
        params = cls.command_control(fcp, args)
        return params[6:]

    @classmethod
    def get_vendor_dependent(cls, fcp, company_ids, deps):
        if len(company_ids) != 3:
            raise ValueError('Invalid array for company ID')
        if len(deps) == 0:
            raise ValueError('Invalid data for vendor dependent field')
        args = bytearray()
        args.append(0x01)
        args.append(0xff)
        args.append(0x00)
        for b in company_ids:
            args.append(b)
        for b in deps:
            args.append(b)
        params = cls.command_status(fcp, args)
        return params[6:]


class AvcConnection():
    PLUG_DIRECTION = ('output', 'input')
    SAMPLING_RATES = (32000, 44100, 48000, 88200, 96000, 176400, 192000)

    @classmethod
    def get_unit_plug_info(cls, fcp):
        args = bytearray()
        args.append(0x01)
        args.append(0xff)
        args.append(0x02)   # Plug info
        args.append(0x00)   # Serial Bus Isochronous and External Plug
        args.append(0xff)
        args.append(0xff)
        args.append(0xff)
        args.append(0xff)
        params = AvcGeneral.command_status(fcp, args)
        return {'isoc': {
            'input':    params[4],
            'output':   params[5]},
            'external': {
            'input':    params[6],
            'output':   params[7]}}

    @classmethod
    def get_subunit_plug_info(cls, fcp, subunit_type, subunit_id):
        if subunit_type not in AvcGeneral.SUBUNIT_TYPES:
            raise ValueError('Invalid argument for subunit type')
        if subunit_id > 7:
            raise ValueError('Invalid argument for subunit id')
        args = bytearray()
        args.append(0x01)
        args.append((AvcGeneral.SUBUNIT_TYPES.index(
            subunit_type) << 3) | subunit_id)
        args.append(0x02)
        args.append(0x00)
        args.append(0xff)
        args.append(0xff)
        args.append(0xff)
        args.append(0xff)
        params = AvcGeneral.command_status(fcp, args)
        # Consider that destination is input and source is output.
        return {'input': params[4], 'output': params[5]}

    @classmethod
    def set_plug_signal_format(cls, fcp, direction, plug, rate):
        if plug > 255:
            raise ValueError('Invalid argument for plug number')
        if direction not in AvcConnection.PLUG_DIRECTION:
            raise ValueError('Invalid argument for plug direction')
        if rate not in AvcConnection.SAMPLING_RATES:
            raise ValueError('Invalid argument for sampling rate')
        args = bytearray()
        args.append(0x00)
        args.append(0xff)
        args.append(0x18 + AvcConnection.PLUG_DIRECTION.index(direction))
        args.append(plug)
        args.append(0x90)
        args.append(AvcConnection.SAMPLING_RATES.index(rate))
        args.append(0xff)
        args.append(0xff)
        AvcGeneral.command_control(fcp, args)

    @classmethod
    def get_plug_signal_format(cls, fcp, direction, plug):
        if plug > 255:
            raise ValueError('Invalid argument for plug number')
        if direction not in AvcConnection.PLUG_DIRECTION:
            raise ValueError('Invalid argument for plug direction')
        args = bytearray()
        args.append(0x01)
        args.append(0xff)
        args.append(0x18 + AvcConnection.PLUG_DIRECTION.index(direction))
        args.append(plug)
        args.append(0xff)
        args.append(0xff)
        args.append(0xff)
        args.append(0xff)
        params = AvcGeneral.command_status(fcp, args)
        param = params[5] & 0x07
        if param > len(AvcConnection.SAMPLING_RATES):
            raise OSError
        return AvcConnection.SAMPLING_RATES[param]

    @classmethod
    def ask_plug_signal_format(cls, fcp, direction, plug, rate):
        if plug > 255:
            raise ValueError('Invalid argument for plug number')
        if direction not in AvcConnection.PLUG_DIRECTION:
            raise ValueError('Invalid argument for plug direction')
        if rate not in AvcConnection.SAMPLING_RATES:
            raise ValueError('Invalid argument for sampling rate')
        args = bytearray()
        args.append(0x02)
        args.append(0xff)
        args.append(0x18 + AvcConnection.PLUG_DIRECTION.index(direction))
        args.append(plug)
        args.append(0x90)
        args.append(AvcConnection.SAMPLING_RATES.index(rate))
        args.append(0xff)
        args.append(0xff)
        try:
            AvcGeneral.command_inquire(fcp, args)
        except OSError:
            return False
        return True



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.bebob.phase_go_protocol_abstract import PhaseGoProtocolAbstract

from hinawa_utils.ta1394.audio import AvcAudio

__all__ = ['PhaseGoProtocolOpt']


class PhaseGoProtocolOpt(PhaseGoProtocolAbstract):
    __MIXER_OUTPUT_FB = 2
    __OUTPUT_LABELS = {
        'analog-1/2':       1,
        'analog-3/4':       2,
        'digital-1/2':      3,
    }

    __ANALOG_OUTPUT_TARGETS = ('analog-1/2', 'analog-3/4')
    __ANALOG_OUTPUT_FB = 1

    def __init__(self, fcp):
        super().__init__(fcp, self.__MIXER_OUTPUT_FB, self.__OUTPUT_LABELS)

    def get_analog_output_labels(self):
        return self.__ANALOG_OUTPUT_TARGETS

    def __check_analog_output_channel(self, target, ch):
        if target not in self.__ANALOG_OUTPUT_TARGETS:
            raise ValueError('Invalid argument for stereo pair')
        if ch not in (1, 2):
            raise ValueError('Invalid argument for channel number')
        return self.__ANALOG_OUTPUT_FB, ch

    def set_analog_output_volume(self, target, ch, db):
        fb, ch = self.__check_analog_output_channel(target, ch)
        data = AvcAudio.build_data_from_db(db)
        AvcAudio.set_feature_volume_state(self.fcp, 0, 'current', fb, ch,
                                          data)

    def get_analog_output_volume(self, target, ch):
        fb, ch = self.__check_analog_output_channel(target, ch)
        data = AvcAudio.get_feature_volume_state(self.fcp, 0, 'current', fb,
                                                 ch)
        return AvcAudio.parse_data_to_db(data)

    def set_analog_output_mute(self, target, ch, enable):
        fb, ch = self.__check_analog_output_channel(target, ch)
        AvcAudio.set_feature_mute_state(self.fcp, 0, 'current', fb, ch,
                                        enable)

    def get_analog_output_mute(self, target, ch):
        fb, ch = self.__check_analog_output_channel(target, ch)
        return AvcAudio.get_feature_mute_state(self.fcp, 0, 'current', fb, ch)



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from struct import unpack
from enum import Enum

__all__ = ['Ieee1212ConfigRomLexer']


class EntryType(Enum):
    IMMEDIATE = 0x00
    CSR_OFFSET = 0x01
    LEAF = 0x02
    DIRECTORY = 0x03

    @classmethod
    def check_value(cls, value):
        return value in (item.value for item in cls)

    def __repr__(self):
        return "'" + self.name + "'"


class Ieee1212ConfigRomLexer():
    @classmethod
    def detect_entries(cls, data):
        entries = {}

        bus_info_length = cls._detect_bus_info_length(data)
        entries['bus-info'] = data[4:4 + bus_info_length]
        data = data[4 + bus_info_length:]

        entries['root-directory'] = cls._detect_directory_entries(data)

        return entries

    @classmethod
    def _detect_bus_info_length(cls, data):
        bus_info_quadlet_count = data[0]
        crc_quadlet_count = data[1]
        crc = unpack('>H', data[2:4])[0]
        return bus_info_quadlet_count * 4

    @classmethod
    def _detect_leaf_length(cls, data):
        quadlet_count = unpack('>H', data[0:2])[0]
        crc = unpack('>H', data[2:4])[0]
        return quadlet_count * 4

    @classmethod
    def _detect_directory_length(cls, data):
        quadlet_count = unpack('>H', data[0:2])[0]
        crc = unpack('>H', data[2:4])[0]
        return quadlet_count * 4

    @classmethod
    def _detect_immediate(cls, key, value, data):
        return value

    @classmethod
    def _detect_csr_offset(cls, key, value, data):
        return 0xfffff0000000 + value * 4

    @classmethod
    def _detect_leaf(cls, key, value, data):
        offset = value * 4
        data = data[offset:]
        length = cls._detect_leaf_length(data)
        return data[4:4 + length]

    @classmethod
    def _detect_directory(cls, key, value, data):
        data = data[value * 4:]
        return cls._detect_directory_entries(data)

    @classmethod
    def _detect_directory_entries(cls, data):
        #
        # Table 7 - Directory entry types
        #
        TYPE_HANDLES = {
            EntryType.IMMEDIATE:   cls._detect_immediate,
            EntryType.CSR_OFFSET:  cls._detect_csr_offset,
            EntryType.LEAF:        cls._detect_leaf,
            EntryType.DIRECTORY:   cls._detect_directory,
        }
        entries = []

        length = cls._detect_directory_length(data)
        data = data[4:]

        while length > 0:
            type_id = data[0] >> 6
            key_id = data[0] & 0x3f
            value = (data[1] << 16) | (data[2] << 8) | data[3]

            if not EntryType.check_value(type_id):
                raise ValueError('Type {0} is not defined.'.format(type_id))
            type = EntryType(type_id)

            entry = [(key_id, type), TYPE_HANDLES[type](key_id, value, data)]
            entries.append(entry)

            data = data[4:]
            length -= 4

        return entries



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.ta1394.general import AvcGeneral

__all__ = ['AvcStreamFormatInfo']


class AvcStreamFormatInfo():
    PLUG_DIRECTION = ('input', 'output')
    HIERARCHY_ROOTS = ('DVCR', 'Audio&Music', 'BT.601', 'invalid', 'reserved')

    AM_HIERARCHY_LEVEL1S = ('am824', 'audio-pack', 'floating-point',
                            'am824-compound')

    SAMPLING_RATES = (22050, 24000, 32000, 44100, 48000, 96000, 176400,
                      192000, 0, 0, 88200, 0, 0, 0, 0, 0)
    RATE_CONTROLS = ('command-based', 'clock-based', 'not-supported')
    TYPES = ('IEC60958-3',      # 0x00
             'IEC61937-3',
             'IEC61937-4',
             'IEC61937-5',
             'IEC61937-6',
             'IEC61937-7',
             'multi-bit-linear-audio-raw',
             'multi-bit-linear-audio-DVD-audio',
             'one-bit-audio-plain-raw',
             'one-bit-audio-plain-SACD',
             'one-bit-audio-encoded-raw',
             'one-bit-audio-encoded-SACD',
             'high-precision-multi-bit-linear-audio',
             'MIDI-conformant',
             'SMPTE-time-code-comformant',
             'sample-count',    # 0x0f
             'ancillary-data',  # 0x10
             'sync-stream',     # 0x40
             'do-not-care',     # 0xff
             'reserved')        # the others

    @classmethod
    def _parse_format(cls, params):
        if params[0] != 0x90 or params[1] != 0x40:
            raise RuntimeError('Unsupported format')
        fmt = {}
        fmt['sampling-rate'] = cls.SAMPLING_RATES[params[2]]
        fmt['rate-control'] = cls.RATE_CONTROLS[params[3] & 0x03]
        formation = []
        for i in range(params[4]):
            for c in range(params[5 + i * 2]):
                type = params[5 + i * 2 + 1]
                if type <= 0x0f:
                    formation.append(cls.TYPES[type])
                elif type == 0x10:
                    formation.append('ancillary-data')
                elif type == 0x40:
                    formation.append('sync-stream')
                elif type == 0xff:
                    formation.append('do-not-care')
                else:
                    formation.append('reserved')
        fmt['formation'] = formation
        return fmt

    @classmethod
    def _build_format(cls, fmt):
        if fmt['sampling-rate'] not in cls.SAMPLING_RATES:
            raise ValueError('Invalid argument for sampling rate')
        if fmt['rate-control'] not in cls.RATE_CONTROLS:
            raise ValueError('Invalid argument for rate control mode')
        args = bytearray()
        args.append(0x90)
        args.append(0x40)
        args.append(cls.SAMPLING_RATES.index(fmt['sampling-rate']))
        args.append(cls.RATE_CONTROLS.index(fmt['rate-control']))
        args.append(0x00)     # Set it later.
        prev = ''
        num = -1
        for i, formation in enumerate(fmt['formation']):
            if formation not in cls.TYPES:
                raise ValueError('Invalid argument for stream formation type')
            if formation == 'ancillary-data':
                type = 0x10
            elif formation == 'sync-stream':
                type = 0x40
            elif formation == 'do-not-care':
                type = 0xff
            elif formation == 'reserved':
                # Use this value.
                type = 0xfe
            else:
                type = cls.TYPES.index(formation)
            if type != prev or i == len(fmt['formation']) - 1:
                if num > 0:
                    args.append(num + 1)
                    args.append(type)
                    args[4] += 1
                num = 1
            else:
                num += 1
            prev = type
        return args

    @classmethod
    def set_format(cls, fcp, direction, plug, fmt):
        args = bytearray()
        args.append(0x00)   # Control
        args.append(0xff)   # Addressing to unit
        args.append(0xbf)   # Extended stream format information command
        args.append(0xc0)   # SINGLE subfunction
        args.append(cls.PLUG_DIRECTION.index(direction))
        args.append(0x00)
        args.append(0x00)
        args.append(plug)
        args.append(0xff)
        args.append(0xff)
        args.extend(cls._build_format(fmt))
        AvcGeneral.command_control(fcp, args)

    @classmethod
    def get_format(cls, fcp, direction, plug):
        if direction not in cls.PLUG_DIRECTION:
            raise ValueError('Invalid argument for plug direction')
        if plug > 255:
            raise ValueError('Invalid argument for plug number')
        args = bytearray()
        args.append(0x01)
        args.append(0xff)
        args.append(0xbf)
        args.append(0xc0)
        args.append(cls.PLUG_DIRECTION.index(direction))
        args.append(0x00)
        args.append(0x00)
        args.append(plug)
        args.append(0xff)
        args.append(0xff)
        args.append(0xff)
        args.append(0xff)
        params = AvcGeneral.command_status(fcp, args)

        return cls._parse_format(params[10:len(params)])

    @classmethod
    def get_formats(cls, fcp, direction, plug):
        if direction not in cls.PLUG_DIRECTION:
            raise ValueError('Invalid argument for plug direction')
        if plug > 255:
            raise ValueError('Invalid argument for plug number')
        fmts = []
        args = bytearray()
        args.append(0x01)
        args.append(0xff)
        args.append(0xbf)
        args.append(0xc1)
        args.append(cls.PLUG_DIRECTION.index(direction))
        args.append(0x00)
        args.append(0x00)
        args.append(plug)
        args.append(0xff)
        args.append(0xff)
        args.append(0x00)
        args.append(0xff)
        for i in range(255):
            args[10] = i
            try:
                params = AvcGeneral.command_status(fcp, args)
                fmt = cls._parse_format(params[11:])
                fmts.append(fmt)
            except Exception as e:
                break
        return fmts



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from struct import pack, unpack
from math import pow, log10

from hinawa_utils.dg00x.dg00x_unit import Dg00xUnit

__all__ = ['Dg003Unit']


class Dg003Unit(Dg00xUnit):
    __OFFSET_MIXER_SRC = 0x0300
    __MAX_COEFF = 0x1fffffff

    def __init__(self, path):
        super().__init__(path)

    def set_mixer_status(self, enabled):
        data = bytearray(4)
        data[3] = 0x01 if enabled else 0x00
        self._write_transaction(0x0124, data)

    def get_mixer_status(self):
        data = self._read_transaction(0x0124, 4)
        return bool(data[3])

    def get_mixer_src_labels(self):
        labels = []
        for ch in range(1, 9, 2):
            labels.append('Analog-{0}/{1}'.format(ch, ch + 1))
        for ch in range(1, 3, 2):
            labels.append('S/PDIF-{0}/{1}'.format(ch, ch + 1))
        for ch in range(1, 8, 2):
            labels.append('ADAT-{0}/{1}'.format(ch, ch + 1))
        return labels

    def __write_src_pair(self, pair):
        for elem in pair:
            offset = elem[0]
            data = pack('>I', elem[1])
            self._write_transaction(offset, data)

    def __read_src_pair(self, src, ch):
        pair = []

        labels = self.get_mixer_src_labels()
        if src not in labels:
            raise ValueError('Invalid argument for source of mixer.')

        for offset in (0x08 * ch, 0x08 * ch + 0x04):
            offset += self.__OFFSET_MIXER_SRC + labels.index(src) * 0x10
            data = self._read_transaction(offset, 4)
            pair.append([offset, unpack('>I', data)[0]])

        return pair

    def __parse_val_to_db(self, val):
        if val == 0:
            return float('-inf')
        return 20 * log10(val / self.__MAX_COEFF)

    def __build_val_from_db(self, db):
        if db == float('-inf'):
            return 0
        return int(self.__MAX_COEFF * pow(10, db / 20))

    def set_mixer_src_balance(self, src, ch, balance):
        pair = self.__read_src_pair(src, ch)
        total = pair[0][1] + pair[1][1]

        if total >= self.__MAX_COEFF:
            total = self.__MAX_COEFF
        left_ratio = 100 - balance
        right_ratio = balance

        if balance == float('-inf'):
            total = 0
        pair[0][1] = int(total * left_ratio / 100)
        pair[1][1] = int(total * right_ratio / 100)
        self.__write_src_pair(pair)

    def get_mixer_src_balance(self, src, ch):
        pair = self.__read_src_pair(src, ch)
        total = pair[0][1] + pair[1][1]
        if total == 0:
            return float('-inf')
        return 100 * pair[1][1] / total

    def set_mixer_src_gain(self, src, ch, db):
        pair = self.__read_src_pair(src, ch)
        val = self.__build_val_from_db(db)

        # Calculate ratio between left and right.
        total = pair[0][1] + pair[1][1]
        if total == 0:
            if pair[0][0] % 0x10 == 0x08:
                # Even channel.
                ratio_left = 0
                ratio_right = 1
            else:
                # Odd channel.
                ratio_left = 1
                ratio_right = 0
        else:
            ratio_left = pair[0][1] / total
            ratio_right = 1 - ratio_left

        pair[0][1] = int(val * ratio_left)
        pair[1][1] = int(val * ratio_right)
        self.__write_src_pair(pair)

    def get_mixer_src_gain(self, src, ch):
        pair = self.__read_src_pair(src, ch)
        total = pair[0][1] + pair[1][1]
        if total == 0:
            return float('-inf')
        return self.__parse_val_to_db(total)



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.ta1394.general import AvcGeneral
from hinawa_utils.ta1394.streamformat import AvcStreamFormatInfo

import time

__all__ = ['BcoPlugInfo', 'BcoSubunitInfo', 'BcoVendorDependent',
           'BcoStreamFormatInfo']


class BcoPlugInfo():
    ADDR_DIR = ('input', 'output')
    ADDR_MODE = ('unit', 'subunit', 'function-block')
    ADDR_UNIT_TYPE = ('isoc', 'external', 'async')

    PLUG_TYPE = ('IsoStream', 'AsyncStream', 'MIDI', 'Sync', 'Analog',
                 'Digital', 'Clock')
    CH_LOCATION = ('N/A', 'left-front', 'right-front', 'center', 'subwoofer',
                   'left-surround', 'right-surround', 'left-of-center',
                   'right-of-center', 'surround', 'side-left', 'side-right',
                   'top', 'buttom', 'left-front-effect', 'right-front-effect',
                   'no-position')
    PORT_TYPE = ('speaker', 'headphone', 'microphone', 'line', 'spdif',
                 'adat', 'tdif', 'madi', 'analog', 'digital', 'MIDI', 'no-type')

    @classmethod
    def get_unit_addr(cls, addr_dir, addr_unit_type, plug):
        if addr_dir not in cls.ADDR_DIR:
            raise ValueError('Invalid argument for address direction')
        if addr_unit_type not in cls.ADDR_UNIT_TYPE:
            raise ValueError('Invalid argument for address unit type')
        if plug > 255:
            raise ValueError('Invalid argument for plug number')
        addr = bytearray()
        addr.append(cls.ADDR_DIR.index(addr_dir))
        addr.append(cls.ADDR_MODE.index('unit'))
        addr.append(cls.ADDR_UNIT_TYPE.index(addr_unit_type))
        addr.append(plug)
        addr.append(0xff)
        # For my purpose.
        addr.append(0xff)
        return addr

    @classmethod
    def get_subunit_addr(cls, addr_dir, subunit_type, subunit_id, plug):
        if addr_dir not in cls.ADDR_DIR:
            raise ValueError('Invalid argument for address direction')
        if subunit_type not in AvcGeneral.SUBUNIT_TYPES:
            raise ValueError('Invalid argument for address subunit type')
        if subunit_id > 7:
            raise ValueError('Invalid argument for address subunit id')
        if plug > 255:
            raise ValueError('Invalid argument for address plug number')
        addr = bytearray()
        addr.append(cls.ADDR_DIR.index(addr_dir))
        addr.append(cls.ADDR_MODE.index('subunit'))
        addr.append(plug)
        addr.append(0xff)
        addr.append(0xff)
        # For my purpose.
        addr.append((AvcGeneral.SUBUNIT_TYPES.index(subunit_type) << 3) |
                    subunit_id)
        return addr

    @classmethod
    def get_function_block_addr(cls, addr_dir, subunit_type, subunit_id,
                                fb_type, fb_id, plug):
        if addr_dir not in cls.ADDR_DIR:
            raise ValueError('Invalid argument for address direction')
        if subunit_type not in AvcGeneral.SUBUNIT_TYPES:
            raise ValueError('Invalid argument for address subunit type')
        if subunit_id > 7:
            raise ValueError('Invalid argument for address subunit id')
        addr = bytearray()
        addr.append(cls.ADDR_DIR.index(addr_dir))
        addr.append(cls.ADDR_MODE.index('function-block'))
        addr.append(fb_type)
        addr.append(fb_id)
        addr.append(plug)
        # For my purpose.
        addr.append((AvcGeneral.SUBUNIT_TYPES.index(subunit_type) << 3) |
                    subunit_id)
        return addr

    @classmethod
    def build_plug_info(cls, info):
        addr = bytearray()
        if info['dir'] not in cls.ADDR_DIR:
            raise ValueError('Invalid address direction')
        addr.append(cls.ADDR_DIR.index(info['dir']))
        if info['mode'] not in cls.ADDR_MODE:
            raise ValueError('Invalid address mode')
        addr.append(cls.ADDR_MODE.index(info['mode']))
        data = info['data']
        if info['mode'] == 'unit':
            if data['unit-type'] not in cls.ADDR_UNIT_TYPE:
                raise ValueError('Invalid address unit type')
            addr.append(cls.ADDR_UNIT_TYPE.index(data['unit-type']))
            addr.append(data['plug'])
            addr.append(0xff)
            addr.append(0xff)
            addr.append(0xff)
        else:
            if data['subunit-type'] not in AvcGeneral.SUBUNIT_TYPES:
                raise ValueError('Invalid address subunit type')
            addr.append(AvcGeneral.SUBUNIT_TYPES.index(data['subunit-type']))
            addr.append(data['subunit-id'])
            addr.append(0xff)
            addr.append(0xff)
            if info['mode'] == 'function-block':
                addr.append(data['function-block-type'])
                addr.append(data['function-block-id'])
                addr.append(data['plug'])
        return addr

    @classmethod
    def parse_plug_addr(cls, addr):
        info = {}
        if addr[0] == 0xff:
            return info
        if addr[0] >= len(cls.ADDR_DIR):
            raise OSError('Unexpected address direction')
        info['dir'] = cls.ADDR_DIR[addr[0]]
        if addr[1] >= len(cls.ADDR_MODE):
            raise OSError('Unexpected address mode in response')
        info['mode'] = cls.ADDR_MODE[addr[1]]
        data = {}
        if info['mode'] == 'unit':
            if addr[2] >= len(cls.ADDR_UNIT_TYPE):
                raise OSError('Unexpected address unit type in response')
            data['unit-type'] = cls.ADDR_UNIT_TYPE[addr[2]]
            data['plug'] = addr[3]
        else:
            if addr[2] >= len(AvcGeneral.SUBUNIT_TYPES):
                raise OSError('Unexpected address subunit type in response')
            data['subunit-type'] = AvcGeneral.SUBUNIT_TYPES[addr[2]]
            data['subunit-id'] = addr[3]
            if info['mode'] == 'subunit':
                data['plug'] = addr[4]
            if info['mode'] == 'function-block':
                data['function-block-type'] = addr[4]
                data['function-block-id'] = addr[5]
                data['plug'] = addr[6]
        info['data'] = data
        return info

    @classmethod
    def get_plug_type(cls, fcp, addr):
        args = bytearray()
        args.append(0x01)
        args.append(addr[5])
        args.append(0x02)   # Plug info command
        args.append(0xc0)   # Bco plug info subcommand
        args.append(addr[0])
        args.append(addr[1])
        args.append(addr[2])
        args.append(addr[3])
        args.append(addr[4])
        args.append(0x00)   # Info type is 'type'
        args.append(0xff)
        args.append(0xff)
        params = AvcGeneral.command_status(fcp, args)
        if params[10] > len(cls.PLUG_TYPE):
            raise OSError('Unexpected value in response')
        return cls.PLUG_TYPE[params[10]]

    @classmethod
    def get_plug_name(cls, fcp, addr):
        args = bytearray()
        args.append(0x01)
        args.append(addr[5])
        args.append(0x02)   # Plug info command
        args.append(0xc0)   # Bco plug info subcommand
        args.append(addr[0])
        args.append(addr[1])
        args.append(addr[2])
        args.append(addr[3])
        args.append(addr[4])
        args.append(0x01)   # Info type is 'name'
        args.append(0xff)
        args.append(0xff)
        params = AvcGeneral.command_status(fcp, args)
        length = params[10]
        if length == 0:
            return ""
        return params[11:11 + length].decode()

    @classmethod
    def get_plug_channels(cls, fcp, addr):
        args = bytearray()
        args.append(0x01)
        args.append(addr[5])
        args.append(0x02)   # Plug info command
        args.append(0xc0)   # Bco plug info subcommand
        args.append(addr[0])
        args.append(addr[1])
        args.append(addr[2])
        args.append(addr[3])
        args.append(addr[4])
        args.append(0x02)   # Info type is 'the number of channels'
        args.append(0xff)
        args.append(0xff)
        params = AvcGeneral.command_status(fcp, args)
        return params[10]

    @classmethod
    def get_plug_ch_name(cls, fcp, addr, pos):
        args = bytearray()
        args.append(0x01)
        args.append(addr[5])
        args.append(0x02)   # Plug info command
        args.append(0xc0)   # Bco plug info subcommand
        args.append(addr[0])
        args.append(addr[1])
        args.append(addr[2])
        args.append(addr[3])
        args.append(addr[4])
        args.append(0x04)   # Info type is 'channel position name'
        args.append(pos)
        args.append(0xff)
        params = AvcGeneral.command_status(fcp, args)
        length = params[11]
        return params[12:12 + length].decode()

    @classmethod
    def get_plug_clusters(cls, fcp, addr):
        args = bytearray()
        args.append(0x01)
        args.append(addr[5])
        args.append(0x02)   # Plug info command
        args.append(0xc0)   # Bco plug info subcommand
        args.append(addr[0])
        args.append(addr[1])
        args.append(addr[2])
        args.append(addr[3])
        args.append(addr[4])
        args.append(0x03)   # Info type is 'channel position data'
        args.append(0xff)
        args.append(0xff)
        params = AvcGeneral.command_status(fcp, args)
        data = params[10:]
        pos = 0
        clusters = [[] for i in range(data[pos])]
        pos += 1
        for cls in range(len(clusters)):
            num = data[pos]
            pos += 1
            if num == 0:
                break

            clusters[cls] = [[0, 0] for j in range(num)]
            for e in range(len(clusters[cls])):
                clusters[cls][e][0] = data[pos]
                clusters[cls][e][1] = data[pos + 1]
                pos += 2
        return clusters

    @classmethod
    def get_plug_cluster_info(cls, fcp, addr, cluster):
        args = bytearray()
        args.append(0x01)
        args.append(addr[5])
        args.append(0x02)   # Plug info command
        args.append(0xc0)   # Bco plug info subcommand
        args.append(addr[0])
        args.append(addr[1])
        args.append(addr[2])
        args.append(addr[3])
        args.append(addr[4])
        args.append(0x07)   # Info type is 'cluster info'
        args.append(cluster)
        args.append(0xff)
        params = AvcGeneral.command_status(fcp, args)
        length = params[12]
        return params[13:13 + length].decode()

    @classmethod
    def get_plug_input(cls, fcp, addr):
        args = bytearray()
        args.append(0x01)
        args.append(addr[5])
        args.append(0x02)   # Plug info command
        args.append(0xc0)   # Bco plug info subcommand
        args.append(addr[0])
        args.append(addr[1])
        args.append(addr[2])
        args.append(addr[3])
        args.append(addr[4])
        args.append(0x05)   # Info type is 'input data'
        args.append(0xff)
        args.append(0xff)
        params = AvcGeneral.command_status(fcp, args)
        return cls.parse_plug_addr(params[10:])

    @classmethod
    def get_plug_outputs(cls, fcp, addr):
        args = bytearray()
        args.append(0x01)
        args.append(addr[5])
        args.append(0x02)   # Plug info command
        args.append(0xc0)   # Bco plug info subcommand
        args.append(addr[0])
        args.append(addr[1])
        args.append(addr[2])
        args.append(addr[3])
        args.append(addr[4])
        args.append(0x06)   # Info type is 'output data'
        args.append(0xff)
        args.append(0xff)
        params = AvcGeneral.command_status(fcp, args)
        info = []
        plugs = params[10]
        if plugs != 0xff:
            for i in range(plugs):
                addr = cls.parse_plug_addr(params[11 + i * 7:])
                info.append(addr)
        return info


class BcoSubunitInfo():
    FB_PURPOSE = {
        0x00: 'input-gain',
        0x01: 'output-volume',
        0xff: 'nothing-special'
    }

    @classmethod
    def get_subunit_fb_info(cls, fcp, subunit_type, subunit_id, page, fb_type):
        if subunit_type not in AvcGeneral.SUBUNIT_TYPES:
            raise ValueError('Invalid argument for subunit type')
        if subunit_id > 7:
            raise ValueError('Invalid argument for subunit id')
        args = bytearray(0xff for i in range(30))
        args[0] = 0x01
        args[1] = (AvcGeneral.SUBUNIT_TYPES.index(
            subunit_type) << 3) | subunit_id
        args[2] = 0x31
        args[3] = page
        args[4] = 0xff
        try:
            params = AvcGeneral.command_status(fcp, args)
        except Exception as e:
            if str(e) == 'Not implemented':
                return []
            raise
        entries = []
        for i in range(5):
            if params[5 + 5 * i] == 0xff:
                continue
            entry = {}
            entry['type'] = params[5 + 5 * i]
            entry['id'] = params[6 + 5 * i]
            entry['purpose'] = cls.FB_PURPOSE[params[7 + 5 * i]]
            entry['inputs'] = params[8 + 5 * i]
            entry['outputs'] = params[9 + 5 * i]
            entries.append(entry)
        return entries


class BcoVendorDependent():
    SUPPORTED_SPEC = ('con', 'pro')
    ADDR_DIR = ('input', 'output')

    # For IEC 60958-1, a.k.a. 'concumer' or 'S/PDIF'
    SUPPORTED_CON_STATUS = {
        # name, the number of values
        'consumerUse':             1,
        'linearPCM':               1,
        'copyRight':               1,
        'additionalFormatInfo':    1,
        'channelStatusMode':       1,
        'categoryCode':            1,
        'sourceNumber':            1,
        'channelNumber':           1,
        'samplingFrequency':       1,
        'clkAccuracy':             1,
        'maxWordLength':           1,
        'sampleWordLength':        1,
    }

    # For IEC 60958-3, a.k.a 'professional' or 'AES'
    SUPPORTED_PRO_STATUS = {
        # name, the number of values
        'profesionalUse':          1,
        'linearPCM':               1,
        'audioSignalEmphasis':     1,
        'srcSampleFreqLock':       1,
        'samplingFrequency':       1,
        'channelMode':             1,
        'userBitManagement':       1,
        'auxSampleBitsUse':        1,
        'sourceWordLength':        1,
        'multiChannelDesc':        1,
        'referenceSignal':         1,
        'refSigSamplFreq':         1,
        'scalingFlag':             1,
        'channelOriginData':       4,
        'channelDestData':         4,
        'localSampleAddrCode':     4,
        'timeOfDaySampleAddrCode': 4,
        'reliabilityFlags':        1,
        'cyclicRedundancyCheck':   1,
    }

    __CON_SUBCMDS = {
        'consumerUse':             0x00,
        'linearPCM':               0x01,
        'copyRight':               0x02,
        'additionalFormatInfo':    0x03,
        'channelStatusMode':       0x04,
        'categoryCode':            0x05,
        'sourceNumber':            0x06,
        'channelNumber':           0x07,
        'samplingFrequency':       0x08,
        'clkAccuracy':             0x09,
        'maxWordLength':           0x0a,
        'sampleWordLength':        0x0b,
    }

    __PRO_SUBCMDS = {
        'profesionalUse':          0x00,
        'linearPCM':               0x01,
        'audioSignalEmphasis':     0x02,
        'srcSampleFreqLock':       0x03,
        'samplingFrequency':       0x04,
        'channelMode':             0x05,
        'userBitManagement':       0x06,
        'auxSampleBitsUse':        0x07,
        'sourceWordLength':        0x08,
        'multiChannelDesc':        0x09,
        'referenceSignal':         0x0a,
        'refSigSamplFreq':         0x0b,
        'scalingFlag':             0x0c,
        'channelOriginData':       0x0d,
        'channelDestData':         0x0e,
        'localSampleAddrCode':     0x0f,
        'timeOfDaySampleAddrCode': 0x10,
        'reliabilityFlags':        0x11,
        'cyclicRedundancyCheck':   0x12,
    }

    @classmethod
    def set_digital_channel_status(cls, fcp, spec, name, values):
        if spec == 'con':
            attrs = cls.SUPPORTED_CON_STATUS
            subcmds = cls.__CON_SUBCMDS
        elif spec == 'pro':
            attrs = cls.SUPPORTED_PRO_STATUS
            subcmds = cls.__PRO_SUBCMDS
        else:
            raise ValueError('Invalid argument for specification name')
        if name not in attrs:
            raise ValueError('Invalid argument for attribute name')
        if attrs[name] != 1:
            if type(values) != 'list' or len(values) != attrs[name]:
                raise ValueError('Invalid argument for attribute value length')
        args = bytearray(0xff for i in range(10))
        args[0] = 0x00
        args[1] = 0xff
        args[2] = 0x00
        args[3] = cls.SUPPORTED_SPEC.index(spec)
        args[4] = subcmds[name]
        args[5] = attrs[name]
        if attrs[name] == 1:
            args[6] = values
        else:
            for i in range(len(values)):
                args[6 + i] = values[i]
        AvcGeneral.command_control(fcp, args)

    @classmethod
    def get_digital_channel_status(cls, fcp, spec, name):
        if spec == 'con':
            attrs = cls.SUPPORTED_CON_STATUS
            subcmds = cls.__CON_SUBCMDS
        elif spec == 'pro':
            attrs = cls.SUPPORTED_PRO_STATUS
            subcmds = cls.__PRO_SUBCMDS
        else:
            raise ValueError('Invalid argument for specification name')
        if name not in attrs:
            raise ValueError('Invalid argument for attribute name')
        args = bytearray(0xff for i in range(10))
        args[0] = 0x01
        args[1] = 0xff
        args[2] = 0x00
        args[3] = cls.SUPPORTED_SPEC.index(spec)
        args[4] = subcmds[name]
        args[5] = attrs[name]
        params = AvcGeneral.command_status(fcp, args)
        return params[6:6 + attrs[name]]

    @classmethod
    def get_stream_detection(cls, fcp, company_ids, dir, ext_plug):
        if dir not in cls.ADDR_DIR:
            raise ValueError('Invalid argument for address direction')
        if ext_plug >= 255:
            raise ValueError('Invalid argument for external plug number')
        args = bytearray()
        args.append(0x00)
        args.append(cls.ADDR_DIR.index(dir))
        args.append(ext_plug)
        args.append(0xff)
        params = AvcGeneral.get_vendor_dependent(fcp, company_ids, args)
        if params[0] != args[0] or params[1] != args[1] or params[2] != args[2]:
            raise OSError('Unexpected value in response')
        if params[3] == 0x00:
            return False
        return True


class BcoStreamFormatInfo():
    format_types = ('Compound', 'Sync')
    data_types = ('IEC60958-3',     # 0x00
                  'IEC61937-3',
                  'IEC61937-4',
                  'IEC61937-5',
                  'IEC61937-6',
                  'IEC61937-7',
                  'multi-bit-linear-audio-raw',
                  'multi-bit-linear-audio-DVD-audio',
                  'one-bit-audio-plain-raw',
                  'one-bit-audio-plain-SACD',
                  'one-bit-audio-encoded-raw',
                  'one-bit-audio-encoded-SACD',
                  'high-precision-multi-bit-linear-audio',
                  'MIDI-conformant',
                  'sync-stream',    # 0x40
                  'do-not-care',    # 0xff
                  'reserved')       # the others

    @classmethod
    def get_entry_list(cls, fcp, addr):
        fmts = []
        for i in range(0xff):
            # DM1500 tends to cause timeout.
            time.sleep(0.1)
            try:
                args = bytearray()
                args.append(0x01)
                args.append(addr[5])
                args.append(0x2f)   # Bco stream format support
                args.append(0xc1)   # List request
                args.append(addr[0])
                args.append(addr[1])
                args.append(addr[2])
                args.append(addr[3])
                args.append(addr[4])
                args.append(0xff)
                args.append(i)
                args.append(0xff)
                params = AvcGeneral.command_status(fcp, args)
                fmts.append(cls._parse_format(params[11:]))
            except OSError as e:
                if str(e) != 'Rejected':
                    raise
                else:
                    break
        return fmts

    # Two types of sync stream: 0x90/0x00/0x40 and 0x90/0x40 with 'sync-stream'
    @classmethod
    def _parse_format(cls, params):
        fmt = {}
        # Sync stream with stereo raw audio
        if params[0] == 0x90 and params[1] == 0x00 and params[2] == 0x40:
            ctl = params[4] & 0x01
            rate = params[4] >> 8
            fmt['type'] = 'Sync'
            fmt['rate-control'] = AvcStreamFormatInfo.RATE_CONTROLS[ctl]
            fmt['sampling-rate'] = AvcStreamFormatInfo.SAMPLING_RATES[rate]
            fmt['formation'] = ['multi-bit-linear-audio-raw']
            return fmt
        if params[0] != 0x90 or params[1] != 0x40:
            raise RuntimeError('Unsupported format')
        fmt['type'] = 'Compound'
        fmt['sampling-rate'] = AvcStreamFormatInfo.SAMPLING_RATES[params[2]]
        ctl = params[3] & 0x3
        fmt['rate-control'] = AvcStreamFormatInfo.RATE_CONTROLS[ctl]
        formation = []
        for i in range(params[4]):
            for c in range(params[5 + i * 2]):
                type = params[5 + i * 2 + 1]
                if type <= 0x0f:
                    formation.append(cls.data_types[type])
                elif type == 0x40:
                    formation.append('sync-stream')
                elif type == 0xff:
                    formation.append('do-not-care')
                else:
                    formation.append('reserved')
        fmt['formation'] = formation
        return fmt



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.oxfw.oxfw_unit import OxfwUnit
from hinawa_utils.oxfw.tascam_protocol import TascamProtocol

__all__ = ['TascamFireone']


class TascamFireone(OxfwUnit):
    def __init__(self, path):
        super().__init__(path)

        if self.vendor_name != 'TASCAM' or self.model_name != 'FireOne':
            raise ValueError('Unsupported model: {0}, {1}'.format(
                self.vendor_name, self.model_name))

    def get_display_mode_labels(self):
        return TascamProtocol.get_display_mode_labels()

    def set_display_mode(self, mode):
        TascamProtocol.set_display_mode(self.fcp, mode)

    def get_display_mode(self):
        return TascamProtocol.get_display_mode(self.fcp)

    def get_control_mode_labels(self):
        return TascamProtocol.get_control_mode_labels()

    def set_control_mode(self, mode):
        TascamProtocol.set_control_mode(self.fcp, mode)

    def get_control_mode(self):
        return TascamProtocol.get_control_mode(self.fcp)

    def get_input_mode_labels(self):
        return TascamProtocol.get_input_mode_labels()

    def set_input_mode(self, mode):
        TascamProtocol.set_input_mode(self.fcp, mode)

    def get_input_mode(self):
        return TascamProtocol.get_input_mode(self.fcp)

    def get_firmware_version(self):
        return TascamProtocol.get_firmware_version(self.fcp)



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.ieee1394.config_rom_parser import Ieee1394ConfigRomParser

__all__ = ['FFConfigRomParser']


class FFConfigRomParser(Ieee1394ConfigRomParser):
    def parse_rom(self, data):
        entries = super().parse_rom(data)
        return self.__parse_entries(entries['root-directory'])

    def __parse_entries(self, entries):
        info = {}

        if (entries[0] != ['VENDOR', 0x000a35] or
                entries[1][0] != 'NODE_CAPABILITIES' or
                entries[2][0] != 'EUI_64' or
                entries[3][0] != 'UNIT'):
            raise OSError('Invalid format of config ROM.')

        unit_entries = entries[3][1]
        if (unit_entries[0] != ['SPECIFIER_ID', 0x000a35] or
                unit_entries[1][0] != 'VERSION' or
                unit_entries[2] != ['MODEL', 0x101800]):
            raise OSError('Invalid data of config ROM.')

        info['model_id'] = unit_entries[1][1]

        return info



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.ieee1394.config_rom_parser import Ieee1394ConfigRomParser

__all__ = ['Ta1394ConfigRomParser']

# TA Document 1999027 Configuration ROM for AV/C Devices 1.0
# http://1394ta.org/specifications/


class Ta1394ConfigRomParser(Ieee1394ConfigRomParser):
    OUI_1394TA = 0x00a02d
    VERSION_AVC = 0x010001

    def parse_rom(self, data):
        entries = super().parse_rom(data)
        return self.__parse_entries(entries['root-directory'])

    def __parse_entries(self, entries):
        # Recommended layout.
        FIELDS = (
            ('VENDOR',              'vendor-id'),
            ('DESCRIPTOR',          'vendor-name'),
            ('MODEL',               'model-id'),
            ('DESCRIPTOR',          'model-name'),
            ('NODE_CAPABILITIES',   'node-capabilities'),
            ('UNIT',                None),
        )
        info = {}

        for i, field in enumerate(FIELDS):
            entry = entries[i]
            name, alt = field
            if entry[0] != name:
                raise OSError('Invalid format of config ROM.')
            if name == 'UNIT':
                # Check unit.
                entry = entry[1]
                if (entry[0][0] != 'SPECIFIER_ID' or
                    entry[1][0] != 'VERSION' or
                    entry[2] != ['MODEL',      info['model-id']] or
                        entry[3][0] != 'DESCRIPTOR'):
                    raise ValueError('Invalid data of config ROM.')
                info['spec-id'] = entry[0][1]
                info['spec-version'] = entry[1][1]
            else:
                info[alt] = entry[1]

        return info



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from struct import unpack, pack

from hinawa_utils.ta1394.general import AvcGeneral

__all__ = ['AvcAudio']


class AvcAudio():
    ATTRIBUTES = ('resolution', 'minimum', 'maximum', 'default', 'duration',
                  'current', 'move', 'delta')

    ATTRIBUTE_VALUES = {
        'resolution':   0x01,
        'minimum':      0x02,
        'maximum':      0x03,
        'default':      0x04,
        'duration':     0x08,
        'current':      0x10,
        'move':         0x18,
        'delta':        0x19,
    }

    @classmethod
    def set_selector_state(cls, fcp, subunit_id, attr, fb_id, value):
        if subunit_id > 0x07:
            raise ValueError('Invalid argument for subunit ID')
        if attr not in ('current', 'minimum', 'maximum', 'default'):
            raise ValueError('Invalid argument for attribute')
        if fb_id > 255:
            raise ValueError('Invalid argument for function block ID')
        if value > 255:
            raise ValueError('Invalid argument for selector value')
        args = bytearray()
        args.append(0x00)
        args.append(0x08 | (subunit_id & 0x07))
        args.append(0xb8)
        args.append(0x80)   # Selector function block
        args.append(fb_id)
        args.append(cls.ATTRIBUTE_VALUES[attr])
        args.append(0x02)   # Selector length is 2
        args.append(value)
        args.append(0x01)   # Selector control
        AvcGeneral.command_control(fcp, args)

    @classmethod
    def get_selector_state(cls, fcp, subunit_id, attr, fb_id):
        if subunit_id > 0x07:
            raise ValueError('Invalid argument for subunit ID')
        if attr not in ('current', 'minimum', 'maximum', 'default'):
            raise ValueError('Invalid argument for attribute')
        if fb_id > 255:
            raise ValueError('Invalid argument for function block ID')
        args = bytearray()
        args.append(0x01)
        args.append(0x08 | (subunit_id & 0x07))
        args.append(0xb8)
        args.append(0x80)   # Selector function block
        args.append(fb_id)
        args.append(cls.ATTRIBUTE_VALUES[attr])
        args.append(0x02)   # Selector length is 2
        args.append(0xff)
        args.append(0x01)   # Selector control
        params = AvcGeneral.command_status(fcp, args)
        return params[7]

    @classmethod
    def set_feature_mute_state(cls, fcp, subunit_id, attr, fb_id, ch, mute):
        if subunit_id > 0x07:
            raise ValueError('Invalid argument for subunit ID')
        if attr != 'current':
            raise ValueError('Invalid argument for attribute')
        if fb_id > 255:
            raise ValueError('Invalid argument for function block ID')
        if ch > 255:
            raise ValueError('Invalid argument for channel number')
        if mute:
            val = 0x70
        else:
            val = 0x60
        args = bytearray()
        args.append(0x00)
        args.append(0x08 | (subunit_id & 0x07))
        args.append(0xb8)
        args.append(0x81)   # Feature function block
        args.append(fb_id)
        args.append(cls.ATTRIBUTE_VALUES[attr])
        args.append(0x02)   # Selector length is 2
        args.append(ch)
        args.append(0x01)   # Mute control
        args.append(0x01)   # Control data length is 1
        args.append(val)
        AvcGeneral.command_control(fcp, args)

    @classmethod
    def get_feature_mute_state(cls, fcp, subunit_id, attr, fb_id, ch):
        if subunit_id > 0x07:
            raise ValueError('Invalid argument for subunit ID')
        if attr != 'current':
            raise ValueError('Invalid argument for attribute')
        if fb_id > 255:
            raise ValueError('Invalid argument for function block ID')
        if ch > 255:
            raise ValueError('Invalid argument for channel number')
        args = bytearray()
        args.append(0x01)
        args.append(0x08 | (subunit_id & 0x07))
        args.append(0xb8)
        args.append(0x81)   # Feature function block
        args.append(fb_id)
        args.append(cls.ATTRIBUTE_VALUES[attr])
        args.append(0x02)   # Selector length is 2
        args.append(ch)
        args.append(0x01)   # Mute control
        args.append(0x01)   # Control data length is 1
        args.append(0xff)   # Status
        params = AvcGeneral.command_status(fcp, args)
        val = params[10]
        if val == 0x70:
            return True
        elif val == 0x60:
            return False
        else:
            raise OSError('Unexpected value in response')

    @classmethod
    def set_feature_volume_state(cls, fcp, subunit_id, attr, fb_id, ch, data):
        if subunit_id > 0x07:
            raise ValueError('Invalid argument for subunit ID')
        if attr not in cls.ATTRIBUTES:
            raise ValueError('Invalid argument for attribute')
        if fb_id > 255:
            raise ValueError('Invalid argument for function block ID')
        if ch > 255:
            raise ValueError('Invalid argument for channel number')
        if len(data) != 2:
            raise ValueError('Invalid argument for data array')
        args = bytearray()
        args.append(0x00)
        args.append(0x08 | (subunit_id & 0x07))
        args.append(0xb8)
        args.append(0x81)   # Feature function block
        args.append(fb_id)
        args.append(cls.ATTRIBUTE_VALUES[attr])
        args.append(0x02)   # Selector length is 2
        args.append(ch)
        args.append(0x02)   # Volume control
        args.append(0x02)   # Control data length is 2
        args.extend(data)   # Higher and lower parts of volume
        AvcGeneral.command_control(fcp, args)

    @classmethod
    def get_feature_volume_state(cls, fcp, subunit_id, attr, fb_id, ch):
        if subunit_id > 0x07:
            raise ValueError('Invalid argument for subunit ID')
        if attr not in cls.ATTRIBUTES:
            raise ValueError('Invalid argument for attribute')
        if fb_id > 255:
            raise ValueError('Invalid argument for function block ID')
        if ch > 255:
            raise ValueError('Invalid argument for channel number')
        args = bytearray()
        args.append(0x01)
        args.append(0x08 | (subunit_id & 0x07))
        args.append(0xb8)
        args.append(0x81)   # Feature function block
        args.append(fb_id)
        args.append(cls.ATTRIBUTE_VALUES[attr])
        args.append(0x02)   # Selector length is 2
        args.append(ch)
        args.append(0x02)   # Volume control
        args.append(0x02)   # Control data length is 2
        args.append(0xff)   # Higher part of volume
        args.append(0xff)   # Lower part of volume
        params = AvcGeneral.command_status(fcp, args)
        data = params[10:12]
        return data

    @classmethod
    def set_feature_lr_state(cls, fcp, subunit_id, attr, fb_id, ch, data):
        if subunit_id > 0x07:
            raise ValueError('Invalid argument for subunit ID')
        if attr not in cls.ATTRIBUTES:
            raise ValueError('Invalid argument for attribute')
        if fb_id > 255:
            raise ValueError('Invalid argument for function block ID')
        if ch > 255:
            raise ValueError('Invalid argument for channel number')
        if len(data) != 2:
            raise ValueError('Invalid argument for data array')
        args = bytearray()
        args.append(0x00)
        args.append(0x08 | (subunit_id & 0x07))
        args.append(0xb8)
        args.append(0x81)   # Feature function block
        args.append(fb_id)
        args.append(cls.ATTRIBUTE_VALUES[attr])
        args.append(0x02)   # Selector length is 2
        args.append(ch)
        args.append(0x03)   # LR control
        args.append(0x02)   # Control data length is 2
        args.extend(data)   # Higher and lower parts of volume
        AvcGeneral.command_control(fcp, args)

    @classmethod
    def get_feature_lr_state(cls, fcp, subunit_id, attr, fb_id, ch):
        if subunit_id > 0x07:
            raise ValueError('Invalid argument for subunit ID')
        if attr not in cls.ATTRIBUTES:
            raise ValueError('Invalid argument for attribute')
        if fb_id > 255:
            raise ValueError('Invalid argument for function block ID')
        if ch > 255:
            raise ValueError('Invalid argument for channel number')
        args = bytearray()
        args.append(0x01)
        args.append(0x08 | (subunit_id & 0x07))
        args.append(0xb8)
        args.append(0x81)   # Feature function block
        args.append(fb_id)
        args.append(cls.ATTRIBUTE_VALUES[attr])
        args.append(0x02)   # Selector length is 2
        args.append(ch)
        args.append(0x03)   # LR control
        args.append(0x02)   # Control data length is 2
        args.append(0xff)   # Higher part of balance
        args.append(0xff)   # Lower part of balance
        params = AvcGeneral.command_status(fcp, args)
        data = params[10:12]
        return data

    @classmethod
    def set_processing_mixer_state(cls, fcp, subunit_id, attr, fb_id, in_fb,
                                   in_ch, out_ch, data):
        attrs = ('current', 'minimum', 'maximum', 'resolution', 'default')
        if subunit_id > 0x07:
            raise ValueError('Invalid argument for subunit ID')
        if attr not in attrs:
            raise ValueError('Invalid argument for attribute')
        if fb_id > 255:
            raise ValueError('Invalid argument for function block ID')
        if in_fb > 255:
            raise ValueError('Invalid argument for input function block ID')
        if in_ch > 255:
            raise ValueError('Invalid argument for input channel number')
        if out_ch > 255:
            raise ValueError('Invalid argument for output channel number')
        if len(data) != 2:
            raise ValueError('Invalid argument for data array')
        args = bytearray()
        args.append(0x00)
        args.append(0x08 | (subunit_id & 0x07))
        args.append(0xb8)
        args.append(0x82)   # Processing function block
        args.append(fb_id)
        args.append(cls.ATTRIBUTE_VALUES[attr])
        args.append(0x04)   # Selector length is 4
        args.append(in_fb)
        args.append(in_ch)
        args.append(out_ch)
        args.append(0x03)   # Mixer control
        args.append(0x02)   # Control data is 2
        args.extend(data)   # Higher and lower parts of setting
        AvcGeneral.command_control(fcp, args)

    @classmethod
    def get_processing_mixer_state(cls, fcp, subunit_id, attr, fb_id, in_fb,
                                   in_ch, out_ch):
        attrs = ('current', 'minimum', 'maximum', 'resolution', 'default')
        if subunit_id > 0x07:
            raise ValueError('Invalid argument for subunit ID')
        if attr not in attrs:
            raise ValueError('Invalid argument for attribute')
        if fb_id > 255:
            raise ValueError('Invalid argument for function block ID')
        if in_fb > 255:
            raise ValueError('Invalid argument for input function block ID')
        if in_ch > 255:
            raise ValueError('Invalid argument for input channel number')
        if out_ch > 255:
            raise ValueError('Invalid argument for output channel number')
        args = bytearray()
        args.append(0x01)
        args.append(0x08 | (subunit_id & 0x07))
        args.append(0xb8)
        args.append(0x82)   # Processing function block
        args.append(fb_id)
        args.append(cls.ATTRIBUTE_VALUES[attr])
        args.append(0x04)   # Selector length is 4
        args.append(in_fb)
        args.append(in_ch)
        args.append(out_ch)
        args.append(0x03)   # Mixer control
        args.append(0x02)   # Control data is 2
        args.append(0xff)   # Higher part of setting
        args.append(0xff)   # Lower part of setting
        params = AvcGeneral.command_status(fcp, args)
        data = params[12:14]
        return data

    @classmethod
    def set_processing_mixer_state_all(cls, fcp, subunit_id, attr, fb_id, in_fb,
                                       data):
        attrs = ('current', 'minimum', 'maximum', 'resolution', 'default')
        if subunit_id > 0x07:
            raise ValueError('Invalid argument for subunit ID')
        if attr not in attrs:
            raise ValueError('Invalid argument for attribute')
        if fb_id > 255:
            raise ValueError('Invalid argument for function block ID')
        if in_fb > 255:
            raise ValueError('Invalid argument for input function block ID')
        for datum in data:
            if len(data) != 2:
                raise ValueError('Invalid argument for array of data array')
        args = bytearray()
        args.append(0x00)
        args.append(0x08 | (subunit_id & 0x07))
        args.append(0xb8)
        args.append(0x82)   # Processing function block
        args.append(fb_id)
        args.append(cls.ATTRIBUTE_VALUES[attr])
        args.append(0x04)   # Selector length is 4
        args.append(in_fb)
        args.append(0xff)
        args.append(0xff)
        args.append(0x03)       # Mixer control
        args.append(len(data))  # The length of control data
        for datum in data:
            args.extend(datum)
        AvcGeneral.command_control(fcp, args)

    @classmethod
    def get_processing_mixer_state_all(cls, fcp, subunit_id, attr,
                                       fb_id, in_fb):
        attrs = ('current', 'minimum', 'maximum', 'resolution', 'default')
        if subunit_id > 0x07:
            raise ValueError('Invalid argument for subunit ID')
        if attr not in attrs:
            raise ValueError('Invalid argument for attribute')
        if fb_id > 255:
            raise ValueError('Invalid argument for function block ID')
        if in_fb > 255:
            raise ValueError('Invalid argument for input function block ID')
        args = bytearray()
        args.append(0x01)
        args.append(0x08 | (subunit_id & 0x07))
        args.append(0xb8)
        args.append(0x82)   # Processing function block
        args.append(fb_id)
        args.append(cls.ATTRIBUTE_VALUES[attr])
        args.append(0x04)   # Selector length is 4
        args.append(in_fb)
        args.append(0xff)
        args.append(0xff)
        args.append(0x03)   # Mixer control
        args.append(0xff)   # The length of control data in response
        params = AvcGeneral.command_status(fcp, args)
        count = params[11] // 2
        data = []
        params = params[12:]
        for i in range(0, count, 2):
            data.append(params[i * 2:i * 2 + 1])
        return data

    # MEMO: 0x8000 represents negative infinite. 0x7fff is invalid. However,
    # in this method, they're used to represent minimum/maximum value.
    @classmethod
    def parse_data_to_db(cls, data):
        if data[0] == 0x80 and data[1] == 0x00:
            return -128.0
        elif data[0] == 0x7f and data[1] == 0xff:
            return 128.0
        else:
            return unpack('>h', data)[0] * 128 / 0x7fff

    @classmethod
    def build_data_from_db(cls, db):
        if db == 128.0:
            return (0x7f, 0xff)
        elif db == -128.0:
            return (0x80, 0x00)
        else:
            return pack('>h', int(0x7fff * db / 128))



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.ieee1394.config_rom_parser import Ieee1394ConfigRomParser

__all__ = ['MotuConfigRomParser']


class MotuConfigRomParser(Ieee1394ConfigRomParser):
    __OUI_MOTU = 0x0001f2

    def parse_rom(self, data):
        entries = super().parse_rom(data)
        return self.__parse_entries(entries['root-directory'])

    def __parse_entries(self, entries):
        FIELDS = (
            ('VENDOR',              'vendor-id'),
            ('NODE_CAPABILITIES',   'node-capabilities'),
            ('UNIT',                None),
            ('EUI_64',              'guid'),
        )
        info = {}

        for i, field in enumerate(FIELDS):
            entry = entries[i]
            name, alt = field
            if entry[0] != name:
                raise OSError('Invalid format of config ROM.')
            if name == 'UNIT':
                entry = entry[1]
                if (entry[0] != ['SPECIFIER_ID', self.__OUI_MOTU] or
                    entry[1][0] != 'VERSION' or
                        entry[2][0] != 'MODEL'):
                    raise ValueError('Invalid data of unit directory.')
                info['version'] = entry[1][1]
                info['model-id'] = entry[2][1]
            else:
                info[alt] = entry[1]

        return info



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from abc import ABCMeta, abstractmethod

import gi
gi.require_version('Hinawa', '3.0')
from gi.repository import Hinawa

__all__ = ['MotuProtocolAbstract']


class MotuProtocolAbstract(metaclass=ABCMeta):
    BASE_ADDR = 0xfffff0000000

    SUPPORTED_SAMPLING_RATES_X1 = (44100, 48000)
    SUPPORTED_SAMPLING_RATES_X2 = (44100, 48000, 88200, 96000)
    SUPPORTED_SAMPLING_RATES_X4 = (44100, 48000, 88200, 96000, 176400, 192000)
    SUPPORTED_OPT_IFACE_MODES = ('None', 'S/PDIF', 'ADAT')
    SUPPORTED_OPT_IFACE_DIRECTIONS = ('in', 'out')
    SUPPORTED_OPT_IFACE_INDEXES = ('A', 'B')

    CLOCK_INTERNAL = 'Internal'
    CLOCK_ADAT_ON_DSUB = 'ADAT on Dsub-9pin interface'
    CLOCK_ADAT_ON_OPT = 'ADAT on optical interface'
    CLOCK_ADAT_ON_OPT_A = 'ADAT on optical interface A'
    CLOCK_ADAT_ON_OPT_B = 'ADAT on optical interface B'
    CLOCK_SPDIF_ON_COAX = 'S/PDIF on coaxial interface'
    CLOCK_SPDIF_ON_OPT = 'S/PDIF on optical interface'
    CLOCK_SPDIF_ON_OPT_A = 'S/PDIF on optical interface A'
    CLOCK_SPDIF_ON_OPT_B = 'S/PDIF on optical interface B'
    CLOCK_AESEBU_XLR = 'AES/EBU on XLR interface'
    CLOCK_WORD_ON_BNC = 'Word clock on BNC interface'
    CLOCK_UNKNOWN = 'Unknown'

    def __init__(self, unit, debug):
        self._unit = unit
        self._debug = bool(debug)

    def read(self, offset, size):
        req = Hinawa.FwReq()
        if size == 4:
            tcode = Hinawa.FwTcode.READ_QUADLET_REQUEST
        else:
            tcode = Hinawa.FwTcode.READ_BLOCK_REQUEST
        addr = self.BASE_ADDR + offset
        frames = bytearray(size)
        frames = req.transaction(self._unit.get_node(), tcode, addr, size,
                                 frames)
        if self._debug:
            print('    read: {0:012x}:'.format(addr))
            for i, frame in enumerate(frames):
                print('        {0:04x}: {1:02x}'.format(offset + i, frame))
        return bytearray(frames)

    def write(self, offset, frames):
        req = Hinawa.FwReq(timeout=100)
        if len(frames) == 4:
            tcode = Hinawa.FwTcode.WRITE_QUADLET_REQUEST
        else:
            tcode = Hinawa.FwTcode.WRITE_BLOCK_REQUEST
        addr = self.BASE_ADDR + offset
        if self._debug:
            print('    write: {0:012x}:'.format(addr))
            for i, frame in enumerate(frames):
                print('        {0:04x}: {1:02x}'.format(offset + i, frame))

        req.transaction(self.get_node(), tcode, addr, len(frames), frames)

    @abstractmethod
    def get_supported_sampling_rates(self):
        pass

    @abstractmethod
    def get_sampling_rate(self):
        pass

    @abstractmethod
    def set_sampling_rate(self, rate):
        pass

    @abstractmethod
    def get_supported_clock_sources(self):
        pass

    @abstractmethod
    def get_clock_source(self):
        pass

    @abstractmethod
    def set_clock_source(self, source):
        pass

    def get_supported_opt_iface_modes(self):
        return self.SUPPORTED_OPT_IFACE_MODES

    def get_supported_opt_iface_directions(self):
        return self.SUPPORTED_OPT_IFACE_DIRECTIONS

    @abstractmethod
    def get_supported_opt_iface_indexes(self):
        pass

    @abstractmethod
    def get_opt_iface_mode(self, direction, index):
        pass

    @abstractmethod
    def set_opt_iface_mode(self, direction, index, mode):
        pass



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.ieee1394.config_rom_parser import Ieee1394ConfigRomParser

__all__ = ['BebobConfigRomParser']


class BebobConfigRomParser(Ieee1394ConfigRomParser):
    def parse_rom(self, data):
        entries = super().parse_rom(data)
        return self.__parse_entries(entries['root-directory'])

    def __parse_entries(self, entries):
        # Typical layout.
        FIELDS = (
            ('HARDWARE_VERSION',    'hardware-version'),
            ('NODE_CAPABILITIES',   'node-capabilities'),
            ('VENDOR',              'vendor-id'),
            ('DESCRIPTOR',          'vendor-name'),
            ('MODEL',               'model-id'),
            ('DESCRIPTOR',          'model-name'),
            # There are entries of VERSION, UNIT and VENDOR_DEPENDENT keys, but
            # model-specific.
        )
        info = {}

        for i, entry in enumerate(entries):
            if i < len(FIELDS):
                name, alt = FIELDS[i]
                if entry[0] != name:
                    raise OSError('Invalid format of config ROM.')
                info[alt] = entry[1]
            elif entry[0] == 'VERSION':
                info['version'] = entry[1]
            elif entry[0] == 'UNIT':
                items = entry[1]
                # Check unit.
                if (items[0] != ['SPECIFIER_ID', 0x00a02d] or
                        items[1][0] != 'VERSION' or
                        items[2] != ['MODEL', info['model-id']] or
                        items[3] != ['DESCRIPTOR', info['model-name']]):
                    raise ValueError('Invalid data of unit directory.')
                info['unit-version'] = items[1][1]
            elif entry[0] == 'DEPENDENT_INFO':
                items = entry[1]
                if (items[0][0] != 'SPECIFIER_ID' or
                        items[1][0] != 'VERSION' or
                        items[2][0][0] != 0x3a or
                        items[3][0][0] != 0x3b or
                        items[4][0][0] != 0x3c or
                        items[5][0][0] != 0x3d):
                    raise ValueError('Invalid data of dependent information.')
                info['addrs'] = [
                    (items[2][1] << 32) + items[3][1],
                    (items[4][1] << 32) + items[5][1],
                ]

        return info



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from struct import unpack, pack
from time import sleep
from math import log10, pow

__all__ = ['ExtCtlSpace', 'ExtCapsSpace', 'ExtCmdSpace', 'ExtMixerSpace',
           'ExtNewRouterSpace', 'ExtPeakSpace', 'ExtNewStreamConfigSpace',
           'ExtCurrentConfigSpace', 'ExtStandaloneSpace', 'ExtAppSpace']

# '3.1 External control private space'


class ExtCtlSpace():
    _EXT_OFFSET = 0x00200000

    _SECTIONS = {
        'caps':                 0x00,
        'cmd':                  0x08,
        'mixer':                0x10,
        'peak':                 0x18,
        'new-router':           0x20,
        'new-stream-config':    0x28,
        'current-config':       0x30,
        'standalone-config':    0x38,
        'application':          0x40,
    }

    @classmethod
    def write_section(cls, protocol, req, section, offset, data):
        if section not in cls._SECTIONS:
            raise ValueError('Invalid name of section: {0}'.format(section))
        offset += cls._EXT_OFFSET + protocol._ext_layout[section]['offset']
        return protocol.write_transactions(req, offset, data)

    @classmethod
    def read_section(cls, protocol, req, section, offset, length):
        if section not in cls._SECTIONS:
            raise ValueError('Invalid name of section: {0}'.format(section))
        offset += cls._EXT_OFFSET + protocol._ext_layout[section]['offset']
        return protocol.read_transactions(req, offset, length)

    @classmethod
    def detect_layout(cls, protocol, req):
        layout = {}

        data = protocol.read_transactions(req, cls._EXT_OFFSET,
                                          len(cls._SECTIONS) * 8)
        for name, offset in cls._SECTIONS.items():
            layout[name] = {
                'offset': unpack('>I', data[offset: offset + 4])[0] * 4,
                'length': unpack('>I', data[offset + 4: offset + 8])[0] * 4,
            }

        protocol._ext_layout = layout
        return layout

# '3.2 Capability space'


class ExtCapsSpace():
    _OFFSET_ROUTER_CAPS = 0x00
    _OFFSET_MIXER_CAPS = 0x04
    _OFFSET_GENERAL_CAPS = 0x08
    _OFFSET_RESERVED_CAPS = 0x0c

    _ASIC_TYPES = {
        0x00: 'DICE-II',
        0x01: 'TCD-2210',
        0x02: 'TCD-2220',
        0x03: 'Unknown',
    }

    @classmethod
    def _parse_router_caps(cls, data):
        caps = {}

        data = data[cls._OFFSET_ROUTER_CAPS:]
        caps['is-exposed'] = bool(data[3] & 0x01)
        caps['is-readonly'] = bool(data[3] & 0x02)
        caps['is-storable'] = bool(data[3] & 0x04)
        caps['maximum-routes'] = unpack('>H', data[0:2])[0]

        return caps

    @classmethod
    def _parse_mixer_caps(cls, data):
        caps = {}

        data = data[cls._OFFSET_MIXER_CAPS:]
        caps['is-exposed'] = bool(data[3] & 0x01)
        caps['is-readonly'] = bool(data[3] & 0x02)
        caps['is-storable'] = bool(data[3] & 0x04)
        caps['input-device-id'] = (data[3] & 0xf0) >> 4
        caps['output-device-id'] = data[2] & 0x0f
        caps['input-channels'] = data[1]
        caps['output-channels'] = data[0]
        return caps

    @classmethod
    def _parse_general_caps(cls, data):
        caps = {}

        data = data[cls._OFFSET_GENERAL_CAPS:]
        caps['dynamic-stream-conf'] = bool(data[3] & 0x01)
        caps['storage-available'] = bool(data[3] & 0x02)
        caps['peak-available'] = bool(data[3] & 0x04)
        caps['maximum-tx-streams'] = (data[3] & 0xf0) >> 4
        caps['maximum-rx-streams'] = data[2] & 0x0f
        caps['storable-stream-conf'] = bool(data[2] & 0x10)
        asic_type = data[1]
        if asic_type < len(cls._ASIC_TYPES) - 1:
            caps['asic-type'] = cls._ASIC_TYPES[asic_type]
        else:
            caps['asic-type'] = cls._ASIC_TYPES[-1]
        return caps

    @classmethod
    def detect_caps(cls, protocol, req):
        caps = {}

        length = protocol._ext_layout['caps']['length']
        data = ExtCtlSpace.read_section(protocol, req, 'caps', 0, length)

        caps['router'] = cls._parse_router_caps(data)
        caps['mixer'] = cls._parse_mixer_caps(data)
        caps['general'] = cls._parse_general_caps(data)
        caps['reserved'] = data[cls._OFFSET_RESERVED_CAPS:]

        protocol._ext_caps = caps

        return caps

# '3.3 Command space'


class ExtCmdSpace():
    _OFFSET_OPCODE = 0x0000
    _OFFSET_RETURN = 0x0004

    _EXECUTE_FLAG = 0x80

    _RETURN_SUCCESS = 0x00
    _RETURN_FAILURE = 0x01

    _OP_CODES = (
        'noop',
        'load-from-router',
        'load-from-stream-config',
        'load-from-router-stream-config',
        'load-from-storage',
        'load-to-storage',
    )

    _RATE_MODES = {
        'low':      0x01,
        'middle':   0x02,
        'high':     0x04,
    }

    @classmethod
    def initiate(cls, protocol, req, cmd, mode):
        if cmd not in cls._OP_CODES:
            raise ValueError('Invalid argument for command')
        if mode not in cls._RATE_MODES:
            raise ValueError('Invalid argument for rate mode')

        # Check feature capabilities.
        if ((cmd.find('router') > 0 and
                protocol._ext_caps['router']['is-readonly']) or
            (cmd.find('stream-config') > 0 and
                not protocol._ext_caps['general']['dynamic-stream-conf']) or
            (cmd in ('load-from-storage', 'load-to-storage') and
                not protocol._ext_caps['general']['storage-available'])):
            raise RuntimeError('This feature is not available.')

        data = bytearray(4)
        data[0] = cls._EXECUTE_FLAG
        data[1] = cls._RATE_MODES[mode]
        data[3] = cls._OP_CODES.index(cmd)

        ExtCtlSpace.write_section(
            protocol, req, 'cmd', cls._OFFSET_OPCODE, data)

        # Completion is notified as clearing of bit flags in the register.
        count = 10
        while count > 0:
            data = ExtCtlSpace.read_section(protocol, req, 'cmd',
                                            cls._OFFSET_OPCODE, 4)
            if not (data[0] & cls._EXECUTE_FLAG):
                break
            count -= 1
            sleep(0.2)
        else:
            raise IOError('Timeout of command initiation.')

        data = ExtCtlSpace.read_section(protocol, req, 'cmd',
                                        cls._OFFSET_RETURN, 4)
        if data[3] != cls._RETURN_SUCCESS:
            raise IOError('Fail to execute requested operation.')

# '3.4 Mixer space'


class ExtMixerSpace():
    # These are TCD-2200/2210 specification.
    MIXER_IN_MAX_PORTS = {
        'mixer-tx0': 16,
        'mixer-tx1':  2,
    }
    MIXER_OUT_MAX_PORTS = {
        'low':      16,
        'middle':   16,
        'high':     8,
    }
    MAX_COEFF = 0x3fff

    # '5.11 Audio Mixer' in 'TCD22xx Users Guide'.
    @classmethod
    def parse_val_to_db(cls, val):
        if val == 0:
            return float('-inf')
        return 20 * log10(val / cls.MAX_COEFF)

    @classmethod
    def build_val_from_db(cls, db):
        if db > 4:
            raise ValueError('Invalid argument for dB value.')
        if db == float('-inf'):
            return 0
        return int(cls.MAX_COEFF * pow(10, db / 20))

    @classmethod
    def _calcurate_offset(cls, protocol, out_ch, in_ch):
        if out_ch >= protocol._ext_caps['mixer']['output-channels']:
            raise ValueError('Invalid value for output channel')
        if in_ch >= protocol._ext_caps['mixer']['input-channels']:
            raise ValueError('Invalid value for input channel')
        offset = (
            out_ch * protocol._ext_caps['mixer']['input-channels'] + in_ch) * 4
        if offset >= protocol._ext_layout['mixer']['length']:
            raise OSError('Inconsistency between channels and length of space')

        return 4 + offset

    @classmethod
    def read_saturation(cls, protocol, req, mode):
        if not protocol._ext_caps['mixer']['is-exposed']:
            raise IOError('This feature is not available.')

        data = ExtCtlSpace.read_section(protocol, req, 'mixer', 0, 4)
        bits = unpack('>I', data)[0]
        outputs = cls.MIXER_OUT_MAX_PORTS[mode]

        saturations = []
        for i in range(outputs):
            saturations.append(bool(bits & (1 << i)))
        return saturations

    @classmethod
    def write_gain(cls, protocol, req, out_ch, in_ch, val):
        if not protocol._ext_caps['mixer']['is-exposed']:
            raise IOError('This feature is not available.')

        offset = cls._calcurate_offset(protocol, out_ch, in_ch)

        data = bytearray()
        data.append(0x00)
        data.append(0x00)
        data.extend(pack('>H', val))

        return ExtCtlSpace.write_section(protocol, req, 'mixer', offset, data)

    @classmethod
    def read_gain(cls, protocol, req, out_ch, in_ch):
        if not protocol._ext_caps['mixer']['is-exposed']:
            raise IOError('This feature is not available.')

        offset = cls._calcurate_offset(protocol, out_ch, in_ch)

        data = ExtCtlSpace.read_section(protocol, req, 'mixer', offset, 4)

        return unpack('>H', data[2:4])[0]

# '3.6 New router space'


class ExtNewRouterSpace():
    _SRC_BLK_IDS = (
        'aes', 'adat', 'mixer', 'reserved0',
        'ins0', 'ins1', 'reserved1', 'reserved2',
        'reserved3', 'reserved4', 'arm-apr-audio', 'avs0',
        'avs1', 'reserved5', 'reserved6', 'mute',
    )
    _DST_BLK_IDS = (
        'aes', 'adat', 'mixer-tx0', 'mixer-tx1',
        'ins0', 'ins1', 'reserved0', 'reserved1',
        'reserved2', 'reserved3', 'arm-apb-audio', 'avs0',
        'avs1', 'reserved4', 'reserved5', 'reserved6',
    )

    # '5.1.3 ROUTERn_ENTRYm' in 'TCD22xx Users Guide'.
    @classmethod
    def parse_entry_data(cls, data):
        # '5.1.4 Source Block ID’s' in 'TCD22xx Users Guide'.
        entry = {}

        src_blk_id = data[0] >> 4
        if src_blk_id >= len(cls._SRC_BLK_IDS):
            raise IOError('Invalid id for source block in router entry')
        entry['src-blk'] = cls._SRC_BLK_IDS[src_blk_id]
        entry['src-ch'] = data[0] & 0x0f

        dst_blk_id = data[1] >> 4
        if dst_blk_id >= len(cls._DST_BLK_IDS):
            raise IOError('Invalid id for destination block in router entry')
        entry['dst-blk'] = cls._DST_BLK_IDS[dst_blk_id]
        entry['dst-ch'] = data[1] & 0x0f

        return entry

    @classmethod
    def parse_data(cls, protocol, req, section, offset, length):
        entries = []

        data = ExtCtlSpace.read_section(protocol, req, section, offset, 4)
        count = unpack('>I', data[0:4])[0]
        if count >= length // 4:
            count = length // 4

        offset += 4
        data = ExtCtlSpace.read_section(protocol, req, section, offset,
                                        count * 4)
        if count > 0:
            for i in range(0, 4 * count, 4):
                entry = cls.parse_entry_data(data[2:4])
                entry['peak'] = unpack('>H', data[0:2])[0]
                entries.append(entry)
                data = data[4:]

        return entries

    @classmethod
    def _build_entry_data(cls, entry):
        KEYS = ('peak', 'src-blk', 'src-ch', 'dst-blk', 'dst-ch')

        # Check keys.
        for key in KEYS:
            if key not in entry:
                raise ValueError('Invalid argument for entry data.')

        data = bytearray()
        data.extend(pack('>H', entry['peak']))

        src = (cls._SRC_BLK_IDS.index(entry['src-blk']) << 4) | entry['src-ch']
        data.append(src)

        dst = (cls._DST_BLK_IDS.index(entry['dst-blk']) << 4) | entry['dst-ch']
        data.append(dst)

        return data

    @classmethod
    def _build_data(cls, length, entries):
        data = bytearray()
        data.extend(pack('>I', len(entries)))
        for entry in entries:
            data.extend(cls._build_entry_data(entry))
        # Padding with zero byte.
        for i in range(len(data), length, 4):
            data.extend((0x00, 0x00, 0x00, 0x00))
        return data

    @classmethod
    def set_entries(cls, protocol, req, entries):
        if (not protocol._ext_caps['router']['is-exposed'] or
                protocol._ext_caps['router']['is-readonly']):
            raise IOError('This feature is not available.')
        length = protocol._ext_layout['new-router']['length']
        if (len(entries) >= protocol._ext_caps['router']['maximum-routes'] or
                len(entries) >= length // 4):
            raise ValueError('Too much entries.')
        data = cls._build_data(length, entries)
        ExtCtlSpace.write_section(protocol, req, 'new-router', 0, data)

    @classmethod
    def get_entries(cls, protocol, req):
        if not protocol._ext_caps['router']['is-exposed']:
            raise IOError('This feature is not available.')
        length = protocol._ext_layout['new-router']['length']
        return cls.parse_data(protocol, req, 'new-router', 0, length)

# '3.5 Peak space'


class ExtPeakSpace():
    @classmethod
    def get(cls, protocol, req):
        if not protocol._ext_caps['general']['peak-available']:
            raise IOError('This feature is not available.')

        entries = []
        length = protocol._ext_layout['peak']['length']
        data = ExtCtlSpace.read_section(protocol, req, 'peak', 0, length)
        for i in range(0, len(data), 4):
            entry = ExtNewRouterSpace.parse_entry_data(data[2:4])
            entry['peak'] = unpack('>H', data[0:2])[0]
            entries.append(entry)
            data = data[4:]
        return entries

# '3.7 New stream config space'


class ExtNewStreamConfigSpace():
    @classmethod
    def _parse_entry(cls, data):
        stream = {}

        stream['pcm'] = unpack('>I', data[0:4])[0]
        data = data[4:]

        stream['midi'] = unpack('>I', data[0:4])[0]
        data = data[4:]

        letters = bytearray()
        for i in range(0, 256, 4):
            letters.extend(list(reversed(data[0:4])))
            data = data[4:]
        stream['formation'] = letters.decode('utf-8').split('\\')[0:-2]

        stream['ac3'] = data[0]

        return stream

    @classmethod
    def parse_data(cls, protocol, req, section, offset, length):
        confs = {}

        data = ExtCtlSpace.read_section(protocol, req, section, offset, 8)
        tx_count = unpack('>I', data[0:4])[0]
        rx_count = unpack('>I', data[4:8])[0]

        offset += 8
        if tx_count > 0:
            confs['tx'] = []
        for i in range(tx_count):
            data = ExtCtlSpace.read_section(protocol, req, section,
                                            offset + 268 * i, 268)
            conf = cls._parse_entry(data)
            if conf['pcm'] > 0 or conf['midi'] > 0:
                confs['tx'].append(conf)

        offset += 268 * tx_count
        if rx_count > 0:
            confs['rx'] = []
        for i in range(rx_count):
            data = ExtCtlSpace.read_section(protocol, req, section,
                                            offset + 268 * i, 268)
            conf = cls._parse_entry(data)
            if conf['pcm'] > 0 or conf['midi'] > 0:
                confs['rx'].append(conf)

        return confs

    @classmethod
    def set_entries(cls, protocol, req, config):
        if not protocol._ext_caps['general']['dynamic-stream-conf']:
            raise IOError('This feature is not available.')

    @classmethod
    def get_entries(cls, protocol, req):
        offset = protocol._ext_layout['new-stream-config']['offset']
        length = protocol._ext_layout['new-stream-config']['length']
        return ExtNewStreamConfigSpace.parse_data(protocol, req,
                                                  'new-stream-config', offset, length)

# '3.8 Current config space'


class ExtCurrentConfigSpace():
    _RATE_MODES = {
        'low':      0x0000,
        'middle':   0x2000,
        'high':     0x4000,
    }

    @classmethod
    def read_router_config(cls, protocol, req, mode):
        if mode not in cls._RATE_MODES:
            raise ValueError('Invalid argument for rate mode.')

        offset = cls._RATE_MODES[mode]

        return ExtNewRouterSpace.parse_data(protocol, req, 'current-config',
                                            offset, 0x1000)

    @classmethod
    def read_stream_config(cls, protocol, req, mode):
        if mode not in cls._RATE_MODES:
            raise ValueError('Invalid argument for rate mode.')

        offset = cls._RATE_MODES[mode] + 0x1000

        return ExtNewStreamConfigSpace.parse_data(protocol, req,
                                                  'current-config', offset, 0x1000)

# '3.9 Stand alone config space'


class ExtStandaloneSpace():
    _AES_EXT_OPTIONS = {
        'high-rate': {
            'off':  0x00,
            'on':   0x01,
        },
    }
    _ADAT_EXT_OPTIONS = {
        'mode': {
            'normal':   0x00,
            'smux-ii':  0x01,
            'smux-iv':  0x02,
            'auto':     0x03,
        },
    }
    _WC_EXT_OPTIONS = {
        'mode': {
            'normal':   0x00,
            'low':      0x01,
            'middle':   0x02,
            'high':     0x03,
        },
        'mul': 0x3fff,
        'div': 0xffff,
    }
    _STANDALONE_SPACE_OFFSETS = {
        'aes1':         0x04,
        'aes2':         0x04,
        'aes3':         0x04,
        'aes4':         0x04,
        'aes-any':      0x04,
        'adat':         0x08,
        'tdif':         0x08,
        'word-clock':   0x0c,
        'internal':     0x10,
    }

    @classmethod
    def write_clock_source(cls, protocol, req, source):
        if source not in protocol._clock_sources:
            raise ValueError('Invalid argument for clock source.')
        val = {v: k for k, v in protocol.CLOCK_BITS.items()}[source]

        data = ExtCtlSpace.read_section(
            protocol, req, 'standalone-config', 0, 4)
        if val != data[3]:
            data[3] = val
            ExtCtlSpace.write_section(
                protocol, req, 'standalone-config', 0, data)

    @classmethod
    def read_clock_source(cls, protocol, req):
        data = ExtCtlSpace.read_section(
            protocol, req, 'standalone-config', 0, 4)
        val = data[3]
        if (val not in protocol.CLOCK_BITS or
                protocol._clock_source_labels[val] == 'Unused'):
            raise OSError('Unexpected return value for clock source.')

        return protocol.CLOCK_BITS[val]

    @classmethod
    def get_source_param_options(cls, protocol, source):
        OPTIONS = {
            'aes1':         ExtStandaloneSpace._AES_EXT_OPTIONS,
            'aes2':         ExtStandaloneSpace._AES_EXT_OPTIONS,
            'aes3':         ExtStandaloneSpace._AES_EXT_OPTIONS,
            'aes4':         ExtStandaloneSpace._AES_EXT_OPTIONS,
            'aes-any':      ExtStandaloneSpace._AES_EXT_OPTIONS,
            'adat':         ExtStandaloneSpace._ADAT_EXT_OPTIONS,
            'tdif':         ExtStandaloneSpace._ADAT_EXT_OPTIONS,
            'word-clock':   ExtStandaloneSpace._WC_EXT_OPTIONS,
            'arx1':         None,
            'arx2':         None,
            'arx3':         None,
            'arx4':         None,
            'internal':     {'rate': protocol._sampling_rates, }
        }
        if source not in protocol._clock_sources or source not in OPTIONS:
            raise ValueError('Invalid argument for clock source.')

        return OPTIONS[source]

    @classmethod
    def write_clock_source_params(cls, protocol, req, source, params):
        if source not in protocol._clock_sources:
            raise ValueError('Invalid argument for clock source.')
        param_options = cls.get_source_param_options(protocol, source)
        for name, param_option in param_options.items():
            if name not in params:
                raise ValueError('Required parameter not in param argument.')

        if source == 'word-clock':
            val = cls._WC_EXT_OPTIONS['mode'][params['mode']]
            val |= (params['mul'] - 1) << 2
            val |= (params['div'] - 1) << 16
        elif source == 'internal':
            val = {v: k for k, v in protocol.RATE_BITS.items()}[params['rate']]
        else:
            for name, param_option in param_options.items():
                for option, flag in param_option.items():
                    if params[name] == option:
                        val = flag
                        break

        data = pack('>I', val)

        offset = cls._STANDALONE_SPACE_OFFSETS[source]
        ExtCtlSpace.write_section(
            protocol, req, 'standalone-config', offset, data)

    @classmethod
    def read_clock_source_params(cls, protocol, req, source):
        if source not in protocol._clock_sources:
            raise ValueError('Invalid argument for clock source.')
        offset = cls._STANDALONE_SPACE_OFFSETS[source]

        data = ExtCtlSpace.read_section(protocol, req, 'standalone-config',
                                        offset, 4)
        val = unpack('>I', data)[0]

        params = {}
        param_options = cls.get_source_param_options(protocol, source)
        if source == 'word-clock':
            mode = {v: k for k, v in param_options['mode'].items()}[val & 0x03]
            params['mode'] = mode
            params['mul'] = ((val >> 2) & 0x3fff) + 1
            params['div'] = (val >> 16) + 1
        elif source == 'internal':
            index = val & 0x0f
            if index not in protocol.RATE_BITS:
                raise IOError('Unexpected state of internal clock rate.')
            params['rate'] = protocol.RATE_BITS[index]
        else:
            val &= 0x0f
            for name, param_option in param_options.items():
                for option, flag in param_option.items():
                    if val == flag:
                        params[name] = option

        return params

# '3.10 Application space'


class ExtAppSpace():
    @classmethod
    def set(cls, protocol, req, offset, data):
        if len(data) >= protocol._ext_layout['application']['length']:
            raise ValueError('Invalid argument for data.')
        ExtCtlSpace.write_section(protocol, req, 'application', offset, data)

    @classmethod
    def get(cls, protocol, req, offset, length):
        if length >= protocol._ext_layout['application']['length']:
            raise ValueError('Invalid argument for data.')
        return ExtCtlSpace.read_section(protocol, req, 'application', offset,
                                        length)



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from threading import Thread
from struct import unpack

import gi
gi.require_version('GLib', '2.0')
gi.require_version('Hinawa', '3.0')
from gi.repository import GLib, Hinawa

from hinawa_utils.ta1394.general import AvcGeneral, AvcConnection
from hinawa_utils.ta1394.ccm import AvcCcm

from hinawa_utils.bebob.config_rom_parser import BebobConfigRomParser
from hinawa_utils.bebob.extensions import BcoPlugInfo

__all__ = ['BebobUnit']


class BebobUnit(Hinawa.SndUnit):
    REG_INFO = 0xffffc8020000

    def __init__(self, path):
        super().__init__()
        self.open(path)
        if self.get_property('type') != 3:
            raise ValueError('The character device is not for BeBoB unit')

        ctx = GLib.MainContext.new()
        self.create_source().attach(ctx)
        self.__unit_dispatcher = GLib.MainLoop.new(ctx, False)
        self.__unit_th = Thread(target=lambda d: d.run(), args=(self.__unit_dispatcher, ))
        self.__unit_th.start()

        node = self.get_node()
        ctx = GLib.MainContext.new()
        node.create_source().attach(ctx)
        self.__node_dispatcher = GLib.MainLoop.new(ctx, False)
        self.__node_th = Thread(target=lambda d: d.run(), args=(self.__node_dispatcher, ))
        self.__node_th.start()

        parser = BebobConfigRomParser()
        info = parser.parse_rom(self.get_node().get_config_rom())
        self.vendor_id = info['vendor-id']
        self.model_id = info['model-id']

        self.fcp = Hinawa.FwFcp()
        self.fcp.bind(self.get_node())
        self.firmware_info = self._get_firmware_info()

    def release(self):
        self.fcp.unbind()
        self.__unit_dispatcher.quit()
        self.__node_dispatcher.quit()
        self.__unit_th.join()
        self.__node_th.join()

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, trace):
        self.release()

    def _get_firmware_info(self):
        def _get_string_literal(params):
            if 0x00 in params:
                return '00000000'
            return params.decode('US-ASCII')

        def _get_time_literal(params):
            if 0x00 in params:
                return '000000'
            return params.decode('US-ASCII')

        req = Hinawa.FwReq()
        frames = bytearray(104)
        params = req.transaction(self.get_node(),
                    Hinawa.FwTcode.READ_BLOCK_REQUEST, BebobUnit.REG_INFO, 104,
                    frames)

        info = {}
        info['manufacturer'] = _get_string_literal(params[0:8])
        info['protocol-version'] = unpack('<I', params[8:12])[0]
        info['guid'] = (unpack('<Q', params[12:20])[0] << 32) | \
            unpack('<I', params[20:24])[0]
        info['model-id'] = unpack('<I', params[24:28])[0]
        info['model-revision'] = unpack('<I', params[28:32])[0]
        info['software'] = {
            'build-date':   _get_string_literal(params[32:40]),
            'build-time':   _get_time_literal(params[40:46]),
            'id':           unpack('<I', params[48:52])[0],
            'version':      unpack('<I', params[52:56])[0],
            'base-address': unpack('<I', params[56:60])[0],
            'max-size':     unpack('<I', params[60:64])[0],
        }
        info['bootloader'] = {
            'build-date':   _get_string_literal(params[64:72]),
            'build-time':   _get_time_literal(params[72:80]),
        }
        info['debugger'] = {
            'build-date':   _get_string_literal(params[80:88]),
            'build-time':   _get_string_literal(params[88:96]),
            'id':           unpack('<I', params[96:100])[0],
            'version':      unpack('<I', params[100:104])[0],
        }

        return info
